# coding=utf-8
#
# Copyright © 2011, 2018 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from sexps import *

def make_test_case(f_name, ret_type, body):
    """Create a simple optimization test case consisting of a single
    function with the given name, return type, and body.

    Global declarations are automatically created for any undeclared
    variables that are referenced by the function.  All undeclared
    variables are assumed to be floats.
    """
    check_sexp(body)
    declarations = {}
    def make_declarations(sexp, already_declared = ()):
        if isinstance(sexp, list):
            if len(sexp) == 2 and sexp[0] == 'var_ref':
                if sexp[1] not in already_declared:
                    declarations[sexp[1]] = [
                        'declare', ['in'], 'float', sexp[1]]
            elif len(sexp) == 4 and sexp[0] == 'assign':
                assert sexp[2][0] == 'var_ref'
                if sexp[2][1] not in already_declared:
                    declarations[sexp[2][1]] = [
                        'declare', ['out'], 'float', sexp[2][1]]
                make_declarations(sexp[3], already_declared)
            else:
                already_declared = set(already_declared)
                for s in sexp:
                    if isinstance(s, list) and len(s) >= 4 and \
                            s[0] == 'declare':
                        already_declared.add(s[3])
                    else:
                        make_declarations(s, already_declared)
    make_declarations(body)
    return list(declarations.values()) + \
        [['function', f_name, ['signature', ret_type, ['parameters'], body]]]


# The following functions can be used to build expressions.

def const_float(value):
    """Create an expression representing the given floating point value."""
    return ['constant', 'float', ['{0:.6f}'.format(value)]]

def const_bool(value):
    """Create an expression representing the given boolean value.

    If value is not a boolean, it is converted to a boolean.  So, for
    instance, const_bool(1) is equivalent to const_bool(True).
    """
    return ['constant', 'bool', ['{0}'.format(1 if value else 0)]]

def gt_zero(var_name):
    """Create Construct the expression var_name > 0"""
    return ['expression', 'bool', '<', const_float(0), ['var_ref', var_name]]


# The following functions can be used to build complex control flow
# statements.  All of these functions return statement lists (even
# those which only create a single statement), so that statements can
# be sequenced together using the '+' operator.

def return_(value = None):
    """Create a return statement."""
    if value is not None:
        return [['return', value]]
    else:
        return [['return']]

def break_():
    """Create a break statement."""
    return ['break']

def continue_():
    """Create a continue statement."""
    return ['continue']

def simple_if(var_name, then_statements, else_statements = None):
    """Create a statement of the form

    if (var_name > 0.0) {
       <then_statements>
    } else {
       <else_statements>
    }

    else_statements may be omitted.
    """
    if else_statements is None:
        else_statements = []
    check_sexp(then_statements)
    check_sexp(else_statements)
    return [['if', gt_zero(var_name), then_statements, else_statements]]

def loop(statements):
    """Create a loop containing the given statements as its loop
    body.
    """
    check_sexp(statements)
    return [['loop', statements]]

def declare_temp(var_type, var_name):
    """Create a declaration of the form

    (declare (temporary) <var_type> <var_name)
    """
    return [['declare', ['temporary'], var_type, var_name]]

def assign_x(var_name, value):
    """Create a statement that assigns <value> to the variable
    <var_name>.  The assignment uses the mask (x).
    """
    check_sexp(value)
    return [['assign', ['x'], ['var_ref', var_name], value]]

def complex_if(var_prefix, statements):
    """Create a statement of the form

    if (<var_prefix>a > 0.0) {
       if (<var_prefix>b > 0.0) {
          <statements>
       }
    }

    This is useful in testing jump lowering, because if <statements>
    ends in a jump, lower_jumps.cpp won't try to combine this
    construct with the code that follows it, as it might do for a
    simple if.

    All variables used in the if statement are prefixed with
    var_prefix.  This can be used to ensure uniqueness.
    """
    check_sexp(statements)
    return simple_if(var_prefix + 'a', simple_if(var_prefix + 'b', statements))

def declare_execute_flag():
    """Create the statements that lower_jumps.cpp uses to declare and
    initialize the temporary boolean execute_flag.
    """
    return declare_temp('bool', 'execute_flag') + \
        assign_x('execute_flag', const_bool(True))

def declare_return_flag():
    """Create the statements that lower_jumps.cpp uses to declare and
    initialize the temporary boolean return_flag.
    """
    return declare_temp('bool', 'return_flag') + \
        assign_x('return_flag', const_bool(False))

def declare_return_value():
    """Create the statements that lower_jumps.cpp uses to declare and
    initialize the temporary variable return_value.  Assume that
    return_value is a float.
    """
    return declare_temp('float', 'return_value')

def declare_break_flag():
    """Create the statements that lower_jumps.cpp uses to declare and
    initialize the temporary boolean break_flag.
    """
    return declare_temp('bool', 'break_flag') + \
        assign_x('break_flag', const_bool(False))

def lowered_return_simple(value = None):
    """Create the statements that lower_jumps.cpp lowers a return
    statement to, in situations where it does not need to clear the
    execute flag.
    """
    if value:
        result = assign_x('return_value', value)
    else:
        result = []
    return result + assign_x('return_flag', const_bool(True))

def lowered_return(value = None):
    """Create the statements that lower_jumps.cpp lowers a return
    statement to, in situations where it needs to clear the execute
    flag.
    """
    return lowered_return_simple(value) + \
        assign_x('execute_flag', const_bool(False))

def lowered_continue():
    """Create the statement that lower_jumps.cpp lowers a continue
    statement to.
    """
    return assign_x('execute_flag', const_bool(False))

def lowered_break_simple():
    """Create the statement that lower_jumps.cpp lowers a break
    statement to, in situations where it does not need to clear the
    execute flag.
    """
    return assign_x('break_flag', const_bool(True))

def lowered_break():
    """Create the statement that lower_jumps.cpp lowers a break
    statement to, in situations where it needs to clear the execute
    flag.
    """
    return lowered_break_simple() + assign_x('execute_flag', const_bool(False))

def if_execute_flag(statements):
    """Wrap statements in an if test so that they will only execute if
    execute_flag is True.
    """
    check_sexp(statements)
    return [['if', ['var_ref', 'execute_flag'], statements, []]]

def if_return_flag(then_statements, else_statements):
    """Wrap statements in an if test with return_flag as the condition.
    """
    check_sexp(then_statements)
    check_sexp(else_statements)
    return [['if', ['var_ref', 'return_flag'], then_statements, else_statements]]

def if_not_return_flag(statements):
    """Wrap statements in an if test so that they will only execute if
    return_flag is False.
    """
    check_sexp(statements)
    return [['if', ['var_ref', 'return_flag'], [], statements]]

def final_return():
    """Create the return statement that lower_jumps.cpp places at the
    end of a function when lowering returns.
    """
    return [['return', ['var_ref', 'return_value']]]

def final_break():
    """Create the conditional break statement that lower_jumps.cpp
    places at the end of a function when lowering breaks.
    """
    return [['if', ['var_ref', 'break_flag'], break_(), []]]

def bash_quote(*args):
    """Quote the arguments appropriately so that bash will understand
    each argument as a single word.
    """
    def quote_word(word):
        for c in word:
            if not (c.isalpha() or c.isdigit() or c in '@%_-+=:,./'):
                break
        else:
            if not word:
                return "''"
            return word
        return "'{0}'".format(word.replace("'", "'\"'\"'"))
    return ' '.join(quote_word(word) for word in args)

def create_test_case(input_sexp, expected_sexp, test_name,
                     pull_out_jumps=False, lower_sub_return=False,
                     lower_main_return=False, lower_continue=False):
    """Create a test case that verifies that do_lower_jumps transforms
    the given code in the expected way.
    """
    check_sexp(input_sexp)
    check_sexp(expected_sexp)
    input_str = sexp_to_string(sort_decls(input_sexp))
    expected_output = sexp_to_string(sort_decls(expected_sexp)) # XXX: don't stringify this
    optimization = (
        'do_lower_jumps({0:d}, {1:d}, {2:d}, {3:d})'.format(
            pull_out_jumps, lower_sub_return, lower_main_return,
            lower_continue))

    return (test_name, optimization, input_str, expected_output)

def test_lower_returns_main():
    """Test that do_lower_jumps respects the lower_main_return flag in deciding
    whether to lower returns in the main function.
    """
    input_sexp = make_test_case('main', 'void', (
            complex_if('', return_())
            ))
    expected_sexp = make_test_case('main', 'void', (
            declare_execute_flag() +
            declare_return_flag() +
            complex_if('', lowered_return())
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_returns_main_true',
        lower_main_return=True)
    yield create_test_case(
        input_sexp, input_sexp, 'lower_returns_main_false',
        lower_main_return=False)

def test_lower_returns_sub():
    """Test that do_lower_jumps respects the lower_sub_return flag in deciding
    whether to lower returns in subroutines.
    """
    input_sexp = make_test_case('sub', 'void', (
            complex_if('', return_())
            ))
    expected_sexp = make_test_case('sub', 'void', (
            declare_execute_flag() +
            declare_return_flag() +
            complex_if('', lowered_return())
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_returns_sub_true',
        lower_sub_return=True)
    yield create_test_case(
        input_sexp, input_sexp, 'lower_returns_sub_false',
        lower_sub_return=False)

def test_lower_returns_1():
    """Test that a void return at the end of a function is eliminated."""
    input_sexp = make_test_case('main', 'void', (
            assign_x('a', const_float(1)) +
            return_()
            ))
    expected_sexp = make_test_case('main', 'void', (
            assign_x('a', const_float(1))
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_returns_1', lower_main_return=True)

def test_lower_returns_2():
    """Test that lowering is not performed on a non-void return at the end of
    subroutine.
    """
    input_sexp = make_test_case('sub', 'float', (
            assign_x('a', const_float(1)) +
            return_(const_float(1))
            ))
    yield create_test_case(
        input_sexp, input_sexp, 'lower_returns_2', lower_sub_return=True)

def test_lower_returns_3():
    """Test lowering of returns when there is one nested inside a complex
    structure of ifs, and one at the end of a function.

    In this case, the latter return needs to be lowered because it will not be
    at the end of the function once the final return is inserted.
    """
    input_sexp = make_test_case('sub', 'float', (
            complex_if('', return_(const_float(1))) +
            return_(const_float(2))
            ))
    expected_sexp = make_test_case('sub', 'float', (
            declare_execute_flag() +
            declare_return_value() +
            declare_return_flag() +
            complex_if('', lowered_return(const_float(1))) +
            if_execute_flag(lowered_return(const_float(2))) +
            final_return()
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_returns_3', lower_sub_return=True)

def test_lower_returns_4():
    """Test that returns are properly lowered when they occur in both branches
    of an if-statement.
    """
    input_sexp = make_test_case('sub', 'float', (
            simple_if('a', return_(const_float(1)),
                      return_(const_float(2)))
            ))
    expected_sexp = make_test_case('sub', 'float', (
            declare_execute_flag() +
            declare_return_value() +
            declare_return_flag() +
            simple_if('a', lowered_return(const_float(1)),
                      lowered_return(const_float(2))) +
            final_return()
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_returns_4', lower_sub_return=True)

def test_lower_unified_returns():
    """If both branches of an if statement end in a return, and pull_out_jumps
    is True, then those returns should be lifted outside the if and then
    properly lowered.

    Verify that this lowering occurs during the same pass as the lowering of
    other returns by checking that extra temporary variables aren't generated.
    """
    input_sexp = make_test_case('main', 'void', (
            complex_if('a', return_()) +
            simple_if('b', simple_if('c', return_(), return_()))
            ))
    expected_sexp = make_test_case('main', 'void', (
            declare_execute_flag() +
            declare_return_flag() +
            complex_if('a', lowered_return()) +
            if_execute_flag(simple_if('b', (simple_if('c', [], []) +
                                            lowered_return())))
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_unified_returns',
        lower_main_return=True, pull_out_jumps=True)

def test_lower_pulled_out_jump():
    doc_string = """If one branch of an if ends in a jump, and control cannot
    fall out the bottom of the other branch, and pull_out_jumps is
    True, then the jump is lifted outside the if.

    Verify that this lowering occurs during the same pass as the
    lowering of other jumps by checking that extra temporary
    variables aren't generated.
    """
    input_sexp = make_test_case('main', 'void', (
            complex_if('a', return_()) +
            loop(simple_if('b', simple_if('c', break_(), continue_()),
                           return_())) +
            assign_x('d', const_float(1))
            ))
    # Note: optimization produces two other effects: the break
    # gets lifted out of the if statements, and the code after the
    # loop gets guarded so that it only executes if the return
    # flag is clear.
    expected_sexp = make_test_case('main', 'void', (
            declare_execute_flag() +
            declare_return_flag() +
            complex_if('a', lowered_return()) +
            if_execute_flag(
                loop(simple_if('b', simple_if('c', [], continue_()),
                               lowered_return_simple()) +
                     break_()) +

                if_return_flag(assign_x('return_flag', const_bool(1)) +
                               assign_x('execute_flag', const_bool(0)),
                               assign_x('d', const_float(1))))
            ))
    yield create_test_case(
        input_sexp, expected_sexp, 'lower_pulled_out_jump',
        lower_main_return=True, pull_out_jumps=True)


def test_remove_continue_at_end_of_loop():
    """Test that a redundant continue-statement at the end of a loop is
    removed.
    """
    input_sexp = make_test_case('main', 'void', (
            loop(assign_x('a', const_float(1)) +
                 continue_())
            ))
    expected_sexp = make_test_case('main', 'void', (
            loop(assign_x('a', const_float(1)))
            ))
    yield create_test_case(input_sexp, expected_sexp, 'remove_continue_at_end_of_loop')

def test_lower_return_void_at_end_of_loop():
    """Test that a return of void at the end of a loop is properly lowered."""
    input_sexp = make_test_case('main', 'void', (
            loop(assign_x('a', const_float(1)) +
                 return_()) +
            assign_x('b', const_float(2))
            ))
    expected_sexp = make_test_case('main', 'void', (
            declare_execute_flag() +
            declare_return_flag() +
            loop(assign_x('a', const_float(1)) +
                 lowered_return_simple() +
                 break_()) +
            if_return_flag(assign_x('return_flag', const_bool(1)) +
                           assign_x('execute_flag', const_bool(0)),
                           assign_x('b', const_float(2)))
            ))
    yield create_test_case(
        input_sexp, input_sexp, 'return_void_at_end_of_loop_lower_nothing')
    yield create_test_case(
        input_sexp, expected_sexp, 'return_void_at_end_of_loop_lower_return',
        lower_main_return=True)


def test_lower_return_non_void_at_end_of_loop():
    """Test that a non-void return at the end of a loop is properly lowered."""
    input_sexp = make_test_case('sub', 'float', (
            loop(assign_x('a', const_float(1)) +
                 return_(const_float(2))) +
            assign_x('b', const_float(3)) +
            return_(const_float(4))
            ))
    expected_sexp = make_test_case('sub', 'float', (
            declare_execute_flag() +
            declare_return_value() +
            declare_return_flag() +
            loop(assign_x('a', const_float(1)) +
                 lowered_return_simple(const_float(2)) +
                 break_()) +
            if_return_flag(assign_x('return_value', '(var_ref return_value)') +
                           assign_x('return_flag', const_bool(1)) +
                           assign_x('execute_flag', const_bool(0)),
                           assign_x('b', const_float(3)) +
                               lowered_return(const_float(4))) +
            final_return()
            ))
    yield create_test_case(
        input_sexp, input_sexp, 'return_non_void_at_end_of_loop_lower_nothing')
    yield create_test_case(
        input_sexp, expected_sexp,
        'return_non_void_at_end_of_loop_lower_return', lower_sub_return=True)


CASES = [
    test_lower_pulled_out_jump,
    test_lower_return_non_void_at_end_of_loop,
    test_lower_return_void_at_end_of_loop,
    test_lower_returns_1, test_lower_returns_2, test_lower_returns_3,
    test_lower_returns_4, test_lower_returns_main, test_lower_returns_sub,
    test_lower_unified_returns, test_remove_continue_at_end_of_loop,
]



#!/usr/bin/python3

import re
import sys

def main():
    if len(sys.argv) != 3:
        print("Missing arguments: ./radv_check_va.py <bo_history> <64-bit VA>")
        sys.exit(1)

    bo_history = str(sys.argv[1])
    va = int(sys.argv[2], 16)

    va_found = False
    with open(bo_history) as f:
        for line in f:
            p = re.compile('timestamp=(.*), VA=(.*)-(.*), destroyed=(.*), is_virtual=(.*)')
            m = p.match(line)
            if m == None:
                continue

            va_start = int(m.group(2), 16)
            va_end = int(m.group(3), 16)

            # Check if the given VA was ever valid and print info.
            if va >= va_start and va < va_end:
                print("VA found: %s" % line, end='')
                va_found = True
    if not va_found:
        print("VA not found!")

if __name__ == '__main__':
    main()



#
# Copyright © 2021 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import argparse
import sys

#
# Tracepoint definitions:
#
def define_tracepoints(args):
    from u_trace import Header, HeaderScope
    from u_trace import ForwardDecl
    from u_trace import Tracepoint
    from u_trace import TracepointArg as Arg
    from u_trace import TracepointArgStruct as ArgStruct

    Header('intel_driver_ds.h', scope=HeaderScope.SOURCE)
    Header('blorp/blorp_priv.h', scope=HeaderScope.HEADER)
    Header('ds/intel_driver_ds.h', scope=HeaderScope.HEADER)

    def begin_end_tp(name, tp_args=[], tp_struct=None, tp_print=None, end_pipelined=True):
        Tracepoint('intel_begin_{0}'.format(name),
                   tp_perfetto='intel_ds_begin_{0}'.format(name))
        Tracepoint('intel_end_{0}'.format(name),
                   args=tp_args,
                   tp_struct=tp_struct,
                   tp_perfetto='intel_ds_end_{0}'.format(name),
                   tp_print=tp_print,
                   end_of_pipe=end_pipelined)


    begin_end_tp('batch',
                 tp_args=[Arg(type='uint8_t', var='name', c_format='%hhu'),],
                 end_pipelined=False)

    begin_end_tp('cmd_buffer',
                 tp_args=[Arg(type='uint8_t', var='level', c_format='%hhu'),],
                 end_pipelined=False)

    begin_end_tp('xfb',
                 end_pipelined=False)

    begin_end_tp('render_pass',
                 tp_args=[Arg(type='uint16_t', var='width', c_format='%hu'),
                          Arg(type='uint16_t', var='height', c_format='%hu'),
                          Arg(type='uint8_t', var='att_count', c_format='%hhu'),
                          Arg(type='uint8_t', var='msaa', c_format='%hhu'),])

    begin_end_tp('dyn_render_pass',
                 tp_args=[Arg(type='uint16_t', var='width', c_format='%hu'),
                          Arg(type='uint16_t', var='height', c_format='%hu'),
                          Arg(type='uint8_t', var='att_count', c_format='%hhu'),
                          Arg(type='uint8_t', var='msaa', c_format='%hhu'),
                          Arg(type='uint8_t', var='suspend', c_format='%hhu'),
                          Arg(type='uint8_t', var='resume', c_format='%hhu'),])

    begin_end_tp('blorp',
                 tp_args=[Arg(type='enum blorp_op', name='op', var='op', c_format='%s', to_prim_type='blorp_op_to_name({})'),
                          Arg(type='uint32_t', name='width', var='width', c_format='%u'),
                          Arg(type='uint32_t', name='height', var='height', c_format='%u'),
                          Arg(type='uint32_t', name='samples', var='samples', c_format='%u'),
                          Arg(type='enum blorp_shader_pipeline', name='blorp_pipe', var='shader_pipe', c_format='%s', to_prim_type='blorp_shader_pipeline_to_name({})'),
                          Arg(type='enum isl_format', name='dst_fmt', var='dst_fmt', c_format='%s', to_prim_type='isl_format_get_short_name({})'),
                          Arg(type='enum isl_format', name='src_fmt', var='src_fmt', c_format='%s', to_prim_type='isl_format_get_short_name({})'),
                          ])

    begin_end_tp('draw',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u')])
    begin_end_tp('draw_multi',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),])
    begin_end_tp('draw_indexed',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),])
    begin_end_tp('draw_indexed_multi',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),])
    begin_end_tp('draw_indirect_byte_count',
                 tp_args=[Arg(type='uint32_t', var='instance_count', c_format='%u'),])
    begin_end_tp('draw_indirect',
                 tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u'),])
    begin_end_tp('draw_indexed_indirect',
                 tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u'),])
    begin_end_tp('draw_indirect_count',
                 tp_args=[Arg(type='uint32_t', var='max_draw_count', c_format='%u'),])
    begin_end_tp('draw_indexed_indirect_count',
                 tp_args=[Arg(type='uint32_t', var='max_draw_count', c_format='%u'),])

    begin_end_tp('draw_mesh',
                 tp_args=[Arg(type='uint32_t', var='group_x', c_format='%u'),
                          Arg(type='uint32_t', var='group_y', c_format='%u'),
                          Arg(type='uint32_t', var='group_z', c_format='%u'),],
                 tp_print=['group=%ux%ux%u', '__entry->group_x', '__entry->group_y', '__entry->group_z'])
    begin_end_tp('draw_mesh_indirect',
                 tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u'),])
    begin_end_tp('draw_mesh_indirect_count',
                 tp_args=[Arg(type='uint32_t', var='max_draw_count', c_format='%u'),])

    begin_end_tp('compute',
                 tp_args=[Arg(type='uint32_t', var='group_x', c_format='%u'),
                          Arg(type='uint32_t', var='group_y', c_format='%u'),
                          Arg(type='uint32_t', var='group_z', c_format='%u'),],
                 tp_print=['group=%ux%ux%u', '__entry->group_x', '__entry->group_y', '__entry->group_z'])

    def flag_bits(args):
        bits = [Arg(type='enum intel_ds_stall_flag', name='flags', var='decode_cb(flags)', c_format='0x%x')]
        for a in args:
            bits.append(Arg(type='bool', name=a[1], var='__entry->flags & INTEL_DS_{0}_BIT'.format(a[0]), c_format='%u'))
        return bits

    def stall_args(args):
        fmt = ''
        exprs = []
        for a in args:
            fmt += '%s'
            exprs.append('(__entry->flags & INTEL_DS_{0}_BIT) ? "+{1}" : ""'.format(a[0], a[1]))
        fmt += ' : %s'
        exprs.append('__entry->reason ? __entry->reason : "unknown"')
        fmt = [fmt]
        fmt += exprs
        return fmt

    stall_flags = [['DEPTH_CACHE_FLUSH',         'depth_flush'],
                   ['DATA_CACHE_FLUSH',          'dc_flush'],
                   ['HDC_PIPELINE_FLUSH',        'hdc_flush'],
                   ['RENDER_TARGET_CACHE_FLUSH', 'rt_flush'],
                   ['TILE_CACHE_FLUSH',          'tile_flush'],
                   ['STATE_CACHE_INVALIDATE',    'state_inval'],
                   ['CONST_CACHE_INVALIDATE',    'const_inval'],
                   ['VF_CACHE_INVALIDATE',       'vf_inval'],
                   ['TEXTURE_CACHE_INVALIDATE',  'tex_inval'],
                   ['INST_CACHE_INVALIDATE',     'ic_inval'],
                   ['STALL_AT_SCOREBOARD',       'pb_stall'],
                   ['DEPTH_STALL',               'depth_stall'],
                   ['CS_STALL',                  'cs_stall']]

    begin_end_tp('stall',
                 tp_args=[ArgStruct(type='uint32_t', var='flags'),
                          ArgStruct(type='intel_ds_stall_cb_t', var='decode_cb'),
                          ArgStruct(type='const char *', var='reason'),],
                 tp_struct=[Arg(type='uint32_t', name='flags', var='decode_cb(flags)', c_format='0x%x'),
                            Arg(type='const char *', name='reason', var='reason', c_format='%s'),],
                 tp_print=stall_args(stall_flags),
                 end_pipelined=False)


def generate_code(args):
    from u_trace import utrace_generate
    from u_trace import utrace_generate_perfetto_utils

    utrace_generate(cpath=args.utrace_src, hpath=args.utrace_hdr,
                    ctx_param='struct intel_ds_device *dev',
                    need_cs_param=False)
    utrace_generate_perfetto_utils(hpath=args.perfetto_hdr)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    parser.add_argument('--utrace-src', required=True)
    parser.add_argument('--utrace-hdr', required=True)
    parser.add_argument('--perfetto-hdr', required=True)
    args = parser.parse_args()
    sys.path.insert(0, args.import_path)
    define_tracepoints(args)
    generate_code(args)


if __name__ == '__main__':
    main()



# Copyright (C) 2012 Intel Corporation
# Copyright (C) 2021 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# This script generates the file api_exec_decl.h, which contains forward
# declarations of all GL functions prefixed with _mesa_*.

import argparse
import collections
import license
import gl_XML
import sys
import apiexec
import textwrap


class PrintCode(gl_XML.gl_print_base):
    def __init__(self):
        super().__init__()

        self.name = 'api_exec_decl_h.py'
        self.license = license.bsd_license_template % (
            'Copyright (C) 2012 Intel Corporation\n'
            'Copyright (C) 2021 Advanced Micro Devices, Inc.',
            'AUTHORS')

    def printBody(self, api):
        print(textwrap.dedent("""\
            #ifndef API_EXEC_DECL_H
            #define API_EXEC_DECL_H

            #include "GL/gl.h"

            #define GL_API GLAPI
            #define GL_APIENTRY GLAPIENTRY
            #include "GLES/gl.h"

            #ifdef __cplusplus
            extern "C" {
            #endif
            """))

        for f in api.functionIterateAll():
            if f.exec_flavor == 'skip':
                continue

            print('{0} GLAPIENTRY _mesa_{1}({2});'.format(
              f.return_type, f.name, f.get_parameter_string()))
            if f.has_no_error_variant:
                print('{0} GLAPIENTRY _mesa_{1}_no_error({2});'.format(
                  f.return_type, f.name, f.get_parameter_string()))
        print('')
        print('#ifdef __cplusplus')
        print('}')
        print('#endif')
        print('')
        print('#endif')


if __name__ == '__main__':
    apiexec.print_glapi_file(PrintCode())



#
# Copyright 2017-2019 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# on the rights to use, copy, modify, merge, publish, distribute, sub
# license, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# THE AUTHOR(S) AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
#
"""
Helper script to merge register database JSON files.

  usage: python3 mergedbs.py [FILES...]

Will merge the given JSON files and output the result on stdout.
"""

from collections import defaultdict
import json
import re
import sys

from regdb import RegisterDatabase, deduplicate_enums, deduplicate_register_types

def main():
    regdb = RegisterDatabase()
    for filename in sys.argv[1:]:
        with open(filename, 'r') as filp:
            regdb.update(RegisterDatabase.from_json(json.load(filp)))

    deduplicate_enums(regdb)
    deduplicate_register_types(regdb)

    print(regdb.encode_json_pretty())


if __name__ == '__main__':
    main()

# kate: space-indent on; indent-width 4; replace-tabs on;



#
# Copyright (C) 2020 Google, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

import argparse
import sys

#
# TODO can we do this with less boilerplate?
#
parser = argparse.ArgumentParser()
parser.add_argument('-p', '--import-path', required=True)
parser.add_argument('-C', '--src', required=True)
parser.add_argument('-H', '--hdr', required=True)
args = parser.parse_args()
sys.path.insert(0, args.import_path)


from u_trace import Header
from u_trace import Tracepoint
from u_trace import TracepointArg
from u_trace import utrace_generate

# List of the default tracepoints enabled. By default tracepoints are enabled,
# set tp_default_enabled=False to disable them by default.
fd_default_tps = []

#
# Tracepoint definitions:
#

Header('util/u_dump.h')
Header('freedreno_batch.h')


def begin_end_tp(name, args=[], tp_struct=None, tp_print=None,
                 tp_default_enabled=True):
    global fd_default_tps
    if tp_default_enabled:
        fd_default_tps.append(name)
    Tracepoint('start_{0}'.format(name),
               toggle_name=name,
               args=args,
               tp_struct=tp_struct,
               tp_perfetto='fd_start_{0}'.format(name),
               tp_print=tp_print)
    Tracepoint('end_{0}'.format(name),
               toggle_name=name,
               tp_perfetto='fd_end_{0}'.format(name))


def singular_tp(name, args=[], tp_struct=None, tp_print=None,
                tp_default_enabled=True):
    global fd_default_tps
    if tp_default_enabled:
        fd_default_tps.append(name)
    Tracepoint(name,
               toggle_name=name,
               args=args,
               tp_struct=tp_struct,
               tp_print=tp_print)

begin_end_tp('state_restore')

singular_tp('flush_batch',
    args=[TracepointArg(type='struct fd_batch *', var='batch',       c_format='%x'),
          TracepointArg(type='uint16_t',          var='cleared',     c_format='%x'),
          TracepointArg(type='uint16_t',          var='gmem_reason', c_format='%x'),
          TracepointArg(type='uint16_t',          var='num_draws',   c_format='%u')],
    tp_print=['%p: cleared=%x, gmem_reason=%x, num_draws=%u', '__entry->batch',
        '__entry->cleared', '__entry->gmem_reason', '__entry->num_draws'],
)

singular_tp('render_gmem',
    args=[TracepointArg(type='uint16_t', var='nbins_x', c_format='%u'),
          TracepointArg(type='uint16_t', var='nbins_y', c_format='%u'),
          TracepointArg(type='uint16_t', var='bin_w',   c_format='%u'),
          TracepointArg(type='uint16_t', var='bin_h',   c_format='%u')],
    tp_print=['%ux%u bins of %ux%u',
        '__entry->nbins_x', '__entry->nbins_y', '__entry->bin_w', '__entry->bin_h'],
)

singular_tp('render_sysmem')

# Note that this doesn't include full information about all of the MRTs
# but seems to roughly match what I see with a blob trace
begin_end_tp('render_pass',
    args=[TracepointArg(type='uint32_t',         var='submit_id',     c_format='%u'),
          TracepointArg(type='enum pipe_format', var='cbuf0_format',  c_format='%s', to_prim_type='util_format_description({})->short_name'),
          TracepointArg(type='enum pipe_format', var='zs_format',     c_format='%s', to_prim_type='util_format_description({})->short_name'),
          TracepointArg(type='uint16_t',         var='width',         c_format='%u'),
          TracepointArg(type='uint16_t',         var='height',        c_format='%u'),
          TracepointArg(type='uint8_t',          var='mrts',          c_format='%u'),
          TracepointArg(type='uint8_t',          var='samples',       c_format='%u'),
          TracepointArg(type='uint16_t',         var='nbins',         c_format='%u'),
          TracepointArg(type='uint16_t',         var='binw',          c_format='%u'),
          TracepointArg(type='uint16_t',         var='binh',          c_format='%u')],
)

begin_end_tp('binning_ib')
begin_end_tp('vsc_overflow_test')
begin_end_tp('prologue')

# For GMEM pass, where this could either be a clear or resolve
begin_end_tp('clear_restore',
    args=[TracepointArg(type='uint16_t', var='fast_cleared', c_format='0x%x')],
    tp_print=['fast_cleared: 0x%x', '__entry->fast_cleared'],
)

begin_end_tp('resolve')

singular_tp('start_tile',
    args=[TracepointArg(type='uint16_t', var='bin_h', c_format='%u'),
          TracepointArg(type='uint16_t', var='yoff',  c_format='%u'),
          TracepointArg(type='uint16_t', var='bin_w', c_format='%u'),
          TracepointArg(type='uint16_t', var='xoff',  c_format='%u')],
    tp_print=['bin_h=%d, yoff=%d, bin_w=%d, xoff=%d',
        '__entry->bin_h', '__entry->yoff', '__entry->bin_w', '__entry->xoff'],
)

begin_end_tp('draw_ib')

begin_end_tp('blit',
    args=[TracepointArg(type='enum pipe_texture_target', var='src_target', c_format='%s', to_prim_type="util_str_tex_target({}, true)"),
          TracepointArg(type='enum pipe_texture_target', var='dst_target', c_format='%s', to_prim_type="util_str_tex_target({}, true)")],
    tp_print=['%s -> %s', 'util_str_tex_target(__entry->src_target, true)',
        'util_str_tex_target(__entry->dst_target, true)'],
)

begin_end_tp('compute')

utrace_generate(cpath=args.src,
                hpath=args.hdr,
                ctx_param='struct pipe_context *pctx',
                trace_toggle_name='fd_gpu_tracepoint',
                trace_toggle_defaults=fd_default_tps)



#!/usr/bin/env python3
# encoding=utf-8
# Copyright 2017-2018 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Script to install megadriver symlinks for meson."""

import argparse
import os


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('megadriver')
    parser.add_argument('libdir')
    parser.add_argument('drivers', nargs='+')
    args = parser.parse_args()

    if os.path.isabs(args.libdir):
        destdir = os.environ.get('DESTDIR')
        if destdir:
            to = os.path.join(destdir, args.libdir[1:])
        else:
            to = args.libdir
    else:
        to = os.path.join(os.environ['MESON_INSTALL_DESTDIR_PREFIX'], args.libdir)

    master = os.path.join(to, os.path.basename(args.megadriver))

    if not os.path.exists(to):
        if os.path.lexists(to):
            os.unlink(to)
        os.makedirs(to)

    for driver in args.drivers:
        abs_driver = os.path.join(to, driver)

        if os.path.lexists(abs_driver):
            os.unlink(abs_driver)
        print('installing {} to {}'.format(args.megadriver, abs_driver))
        os.link(master, abs_driver)

        try:
            ret = os.getcwd()
            os.chdir(to)

            name, ext = os.path.splitext(driver)
            while ext != '.so':
                if os.path.lexists(name):
                    os.unlink(name)
                os.symlink(driver, name)
                name, ext = os.path.splitext(name)
        finally:
            os.chdir(ret)

    # Remove meson-created master .so and symlinks
    os.unlink(master)
    name, ext = os.path.splitext(master)
    while ext != '.so':
        if os.path.lexists(name):
            os.unlink(name)
        name, ext = os.path.splitext(name)


if __name__ == '__main__':
    main()



COPYRIGHT = """\
/*
 * Copyright (C) 2020 Valve Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 */
"""

import argparse
import xml.etree.ElementTree as ET
from mako.template import Template

TEMPLATE  = Template("""\
/* DO NOT EDIT - This file is generated automatically by vtn_generator_ids.py script */

""" + COPYRIGHT + """\
<%
def get_name(generator):
    name = generator.get('tool').lower()
    name = name.replace('-', '')
    name = name.replace(' ', '_')
    name = name.replace('/', '_')
    return name
%>
enum vtn_generator {
% for generator in root.find("./ids[@type='vendor']").findall('id'):
% if 'tool' in generator.attrib:
   vtn_generator_${get_name(generator)} = ${generator.get('value')},
% endif
% endfor
   vtn_generator_max = 0xffff,
};
""")

if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument("xml")
    p.add_argument("out")
    pargs = p.parse_args()

    tree = ET.parse(pargs.xml)
    root = tree.getroot()

    with open(pargs.out, 'w') as f:
        f.write(TEMPLATE.render(root=root))




# (C) Copyright IBM Corporation 2004, 2005
# All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# on the rights to use, copy, modify, merge, publish, distribute, sub
# license, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
# IBM AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# Authors:
#    Ian Romanick <idr@us.ibm.com>

from collections import OrderedDict
from decimal import Decimal
import xml.etree.ElementTree as ET
import re, sys
import os.path
import typeexpr
import static_data


def parse_GL_API( file_name, factory = None ):

    if not factory:
        factory = gl_item_factory()

    api = factory.create_api()
    api.parse_file( file_name )

    # After the XML has been processed, we need to go back and assign
    # dispatch offsets to the functions that request that their offsets
    # be assigned by the scripts.  Typically this means all functions
    # that are not part of the ABI.

    for func in api.functionIterateByCategory():
        if func.assign_offset and func.offset < 0:
            func.offset = api.next_offset;
            api.next_offset += 1

    return api


def is_attr_true( element, name, default = "false" ):
    """Read a name value from an element's attributes.

    The value read from the attribute list must be either 'true' or
    'false'.  If the value is 'false', zero will be returned.  If the
    value is 'true', non-zero will be returned.  An exception will be
    raised for any other value."""

    value = element.get( name, default )
    if value == "true":
        return 1
    elif value == "false":
        return 0
    else:
        raise RuntimeError('Invalid value "%s" for boolean "%s".' % (value, name))


class gl_print_base(object):
    """Base class of all API pretty-printers.

    In the model-view-controller pattern, this is the view.  Any derived
    class will want to over-ride the printBody, printRealHader, and
    printRealFooter methods.  Some derived classes may want to over-ride
    printHeader and printFooter, or even Print (though this is unlikely).
    """

    def __init__(self):
        # Name of the script that is generating the output file.
        # Every derived class should set this to the name of its
        # source file.

        self.name = "a"


        # License on the *generated* source file.  This may differ
        # from the license on the script that is generating the file.
        # Every derived class should set this to some reasonable
        # value.
        #
        # See license.py for an example of a reasonable value.

        self.license = "The license for this file is unspecified."


        # The header_tag is the name of the C preprocessor define
        # used to prevent multiple inclusion.  Typically only
        # generated C header files need this to be set.  Setting it
        # causes code to be generated automatically in printHeader
        # and printFooter.

        self.header_tag = None


        # List of file-private defines that must be undefined at the
        # end of the file.  This can be used in header files to define
        # names for use in the file, then undefine them at the end of
        # the header file.

        self.undef_list = []
        return


    def Print(self, api):
        self.printHeader()
        self.printBody(api)
        self.printFooter()
        return


    def printHeader(self):
        """Print the header associated with all files and call the printRealHeader method."""

        print('/* DO NOT EDIT - This file generated automatically by %s script */' \
                % (self.name))
        print('')
        print('/*')
        print((' * ' + self.license.replace('\n', '\n * ')).replace(' \n', '\n'))
        print(' */')
        print('')
        if self.header_tag:
            print('#if !defined( %s )' % (self.header_tag))
            print('#  define %s' % (self.header_tag))
            print('')
        self.printRealHeader();
        return


    def printFooter(self):
        """Print the header associated with all files and call the printRealFooter method."""

        self.printRealFooter()

        if self.undef_list:
            print('')
            for u in self.undef_list:
                print("#  undef %s" % (u))

        if self.header_tag:
            print('')
            print('#endif /* !defined( %s ) */' % (self.header_tag))


    def printRealHeader(self):
        """Print the "real" header for the created file.

        In the base class, this function is empty.  All derived
        classes should over-ride this function."""
        return


    def printRealFooter(self):
        """Print the "real" footer for the created file.

        In the base class, this function is empty.  All derived
        classes should over-ride this function."""
        return


    def printPure(self):
        """Conditionally define `PURE' function attribute.

        Conditionally defines a preprocessor macro `PURE' that wraps
        GCC's `pure' function attribute.  The conditional code can be
        easilly adapted to other compilers that support a similar
        feature.

        The name is also added to the file's undef_list.
        """
        self.undef_list.append("PURE")
        print("""#  if defined(__GNUC__) || (defined(__SUNPRO_C) && (__SUNPRO_C >= 0x590))
#    define PURE __attribute__((pure))
#  else
#    define PURE
#  endif""")
        return


    def printFastcall(self):
        """Conditionally define `FASTCALL' function attribute.

        Conditionally defines a preprocessor macro `FASTCALL' that
        wraps GCC's `fastcall' function attribute.  The conditional
        code can be easilly adapted to other compilers that support a
        similar feature.

        The name is also added to the file's undef_list.
        """

        self.undef_list.append("FASTCALL")
        print("""#  if defined(__i386__) && defined(__GNUC__) && !defined(__CYGWIN__) && !defined(__MINGW32__)
#    define FASTCALL __attribute__((fastcall))
#  else
#    define FASTCALL
#  endif""")
        return


    def printVisibility(self, S, s):
        """Conditionally define visibility function attribute.

        Conditionally defines a preprocessor macro name S that wraps
        GCC's visibility function attribute.  The visibility used is
        the parameter s.  The conditional code can be easilly adapted
        to other compilers that support a similar feature.

        The name is also added to the file's undef_list.
        """

        self.undef_list.append(S)
        print("""#  if defined(__GNUC__) && !defined(__CYGWIN__) && !defined(__MINGW32__)
#    define %s  __attribute__((visibility("%s")))
#  else
#    define %s
#  endif""" % (S, s, S))
        return


    def printNoinline(self):
        """Conditionally define `NOINLINE' function attribute.

        Conditionally defines a preprocessor macro `NOINLINE' that
        wraps GCC's `noinline' function attribute.  The conditional
        code can be easilly adapted to other compilers that support a
        similar feature.

        The name is also added to the file's undef_list.
        """

        self.undef_list.append("NOINLINE")
        print("""#  if defined(__GNUC__)
#    define NOINLINE __attribute__((noinline))
#  else
#    define NOINLINE
#  endif""")
        return


def real_function_name(element):
    name = element.get( "name" )
    alias = element.get( "alias" )

    if alias:
        return alias
    else:
        return name


def real_category_name(c):
    if re.compile("[1-9][0-9]*[.][0-9]+").match(c):
        return "GL_VERSION_" + c.replace(".", "_")
    else:
        return c


def classify_category(name, number):
    """Based on the category name and number, select a numerical class for it.

    Categories are divided into four classes numbered 0 through 3.  The
    classes are:

            0. Core GL versions, sorted by version number.
            1. ARB extensions, sorted by extension number.
            2. Non-ARB extensions, sorted by extension number.
            3. Un-numbered extensions, sorted by extension name.
    """

    try:
        core_version = float(name)
    except Exception:
        core_version = 0.0

    if core_version > 0.0:
        cat_type = 0
        key = name
    elif name.startswith("GL_ARB_") or name.startswith("GLX_ARB_") or name.startswith("WGL_ARB_"):
        cat_type = 1
        key = int(number)
    else:
        if number != None:
            cat_type = 2
            key = int(number)
        else:
            cat_type = 3
            key = name


    return [cat_type, key]


def create_parameter_string(parameters, include_names):
    """Create a parameter string from a list of gl_parameters."""

    list = []
    for p in parameters:
        if p.is_padding:
            continue

        if include_names:
            list.append( p.string() )
        else:
            list.append( p.type_string() )

    if len(list) == 0: list = ["void"]

    return ", ".join(list)


class gl_item(object):
    def __init__(self, element, context, category):
        self.context = context
        self.name = element.get( "name" )
        self.category = real_category_name( category )

        return


class gl_type( gl_item ):
    def __init__(self, element, context, category):
        gl_item.__init__(self, element, context, category)
        self.size = int( element.get( "size" ), 0 )

        te = typeexpr.type_expression( None )
        tn = typeexpr.type_node()
        tn.size = int( element.get( "size" ), 0 )
        tn.integer = not is_attr_true( element, "float" )
        tn.unsigned = is_attr_true( element, "unsigned" )
        tn.pointer = is_attr_true( element, "pointer" )
        tn.name = "GL" + self.name
        te.set_base_type_node( tn )

        self.type_expr = te
        return


    def get_type_expression(self):
        return self.type_expr


class gl_enum( gl_item ):
    def __init__(self, element, context, category):
        gl_item.__init__(self, element, context, category)
        self.value = int( element.get( "value" ), 0 )

        temp = element.get( "count" )
        if not temp or temp == "?":
            self.default_count = -1
        else:
            try:
                c = int(temp)
            except Exception:
                raise RuntimeError('Invalid count value "%s" for enum "%s" in function "%s" when an integer was expected.' % (temp, self.name, n))

            self.default_count = c

        return


    def priority(self):
        """Calculate a 'priority' for this enum name.

        When an enum is looked up by number, there may be many
        possible names, but only one is the 'prefered' name.  The
        priority is used to select which name is the 'best'.

        Highest precedence is given to core GL name.  ARB extension
        names have the next highest, followed by EXT extension names.
        Vendor extension names are the lowest.
        """

        if self.name.endswith( "_BIT" ):
            bias = 1
        else:
            bias = 0

        if self.category.startswith( "GL_VERSION_" ):
            priority = 0
        elif self.category.startswith( "GL_ARB_" ):
            priority = 2
        elif self.category.startswith( "GL_EXT_" ):
            priority = 4
        else:
            priority = 6

        return priority + bias



class gl_parameter(object):
    def __init__(self, element, context):
        self.name = element.get( "name" )

        ts = element.get( "type" )
        self.type_expr = typeexpr.type_expression( ts, context )

        temp = element.get( "variable_param" )
        if temp:
            self.count_parameter_list = temp.split( ' ' )
        else:
            self.count_parameter_list = []

        # The count tag can be either a numeric string or the name of
        # a variable.  If it is the name of a variable, the int(c)
        # statement will throw an exception, and the except block will
        # take over.

        c = element.get( "count" )
        try: 
            count = int(c)
            self.count = count
            self.counter = None
        except Exception:
            count = 1
            self.count = 0
            self.counter = c

        self.marshal_count = element.get("marshal_count")
        self.count_scale = int(element.get( "count_scale", "1" ))

        elements = (count * self.count_scale)
        if elements == 1:
            elements = 0

        #if ts == "GLdouble":
        #	print '/* stack size -> %s = %u (before)*/' % (self.name, self.type_expr.get_stack_size())
        #	print '/* # elements = %u */' % (elements)
        self.type_expr.set_elements( elements )
        #if ts == "GLdouble":
        #	print '/* stack size -> %s = %u (after) */' % (self.name, self.type_expr.get_stack_size())

        self.is_client_only = is_attr_true( element, 'client_only' )
        self.is_counter     = is_attr_true( element, 'counter' )
        self.is_output      = is_attr_true( element, 'output' )


        # Pixel data has special parameters.

        self.width      = element.get('img_width')
        self.height     = element.get('img_height')
        self.depth      = element.get('img_depth')
        self.extent     = element.get('img_extent')

        self.img_xoff   = element.get('img_xoff')
        self.img_yoff   = element.get('img_yoff')
        self.img_zoff   = element.get('img_zoff')
        self.img_woff   = element.get('img_woff')

        self.img_format = element.get('img_format')
        self.img_type   = element.get('img_type')
        self.img_target = element.get('img_target')

        self.img_pad_dimensions = is_attr_true( element, 'img_pad_dimensions' )
        self.img_null_flag      = is_attr_true( element, 'img_null_flag' )
        self.img_send_null      = is_attr_true( element, 'img_send_null' )

        self.is_padding = is_attr_true( element, 'padding' )
        return


    def compatible(self, other):
        return 1


    def is_array(self):
        return self.is_pointer()


    def is_pointer(self):
        return self.type_expr.is_pointer()


    def is_image(self):
        if self.width:
            return 1
        else:
            return 0


    def is_variable_length(self):
        return len(self.count_parameter_list) or self.counter or self.marshal_count


    def is_64_bit(self):
        count = self.type_expr.get_element_count()
        if count:
            if (self.size() / count) == 8:
                return 1
        else:
            if self.size() == 8:
                return 1

        return 0


    def string(self):
        return self.type_expr.original_string + " " + self.name


    def type_string(self):
        return self.type_expr.original_string


    def get_base_type_string(self):
        return self.type_expr.get_base_name()


    def get_dimensions(self):
        if not self.width:
            return [ 0, "0", "0", "0", "0" ]

        dim = 1
        w = self.width
        h = "1"
        d = "1"
        e = "1"

        if self.height:
            dim = 2
            h = self.height

        if self.depth:
            dim = 3
            d = self.depth

        if self.extent:
            dim = 4
            e = self.extent

        return [ dim, w, h, d, e ]


    def get_stack_size(self):
        return self.type_expr.get_stack_size()


    def size(self):
        if self.is_image():
            return 0
        else:
            return self.type_expr.get_element_size()


    def get_element_count(self):
        c = self.type_expr.get_element_count()
        if c == 0:
            return 1

        return c


    def size_string(self, use_parens = 1, marshal = 0):
        base_size_str = ""

        count = self.get_element_count()
        if count:
            base_size_str = "%d * " % count

        base_size_str += "sizeof(%s)" % ( self.get_base_type_string() )

        if self.counter or self.count_parameter_list or (self.marshal_count and marshal):
            list = [ "compsize" ]

            if self.marshal_count and marshal:
                list = [ self.marshal_count ]
            elif self.counter and self.count_parameter_list:
                list.append( self.counter )
            elif self.counter:
                list = [ self.counter ]

            if self.size() > 1:
                list.append( base_size_str )

            if len(list) > 1 and use_parens :
                return "safe_mul(%s)" % ", ".join(list)
            else:
                return " * ".join(list)

        elif self.is_image():
            return "compsize"
        else:
            return base_size_str


    def format_string(self):
        if self.type_expr.original_string == "GLenum":
            return "0x%x"
        else:
            return self.type_expr.format_string()


class gl_function( gl_item ):
    def __init__(self, element, context):
        self.context = context
        self.name = None

        self.entry_points = []
        self.return_type = "void"
        self.parameters = []
        self.offset = -1
        self.initialized = 0
        self.images = []
        self.exec_flavor = 'mesa'
        self.has_hw_select_variant = False
        self.desktop = True
        self.deprecated = None
        self.has_no_error_variant = False

        # self.api_map[api] is a decimal value indicating the earliest
        # version of the given API in which ANY alias for the function
        # exists.  The map only lists APIs which contain the function
        # in at least one version.  For example, for the ClipPlanex
        # function, self.api_map == { 'es1':
        # Decimal('1.1') }.
        self.api_map = {}

        self.assign_offset = False

        self.static_entry_points = []

        # Track the parameter string (for the function prototype)
        # for each entry-point.  This is done because some functions
        # change their prototype slightly when promoted from extension
        # to ARB extension to core.  glTexImage3DEXT and glTexImage3D
        # are good examples of this.  Scripts that need to generate
        # code for these differing aliases need to real prototype
        # for each entry-point.  Otherwise, they may generate code
        # that won't compile.

        self.entry_point_parameters = {}

        self.process_element( element )

        return


    def process_element(self, element):
        name = element.get( "name" )
        alias = element.get( "alias" )

        # marshal isn't allowed with alias
        assert not alias or not element.get('marshal')
        assert not alias or not element.get('marshal_count')
        assert not alias or not element.get('marshal_sync')
        assert not alias or not element.get('marshal_call_before')
        assert not alias or not element.get('marshal_call_after')

        if name in static_data.functions:
            self.static_entry_points.append(name)

        self.entry_points.append( name )

        for api in ('es1', 'es2'):
            version_str = element.get(api, 'none')
            assert version_str is not None
            if version_str != 'none':
                version_decimal = Decimal(version_str)
                if api not in self.api_map or \
                        version_decimal < self.api_map[api]:
                    self.api_map[api] = version_decimal

        exec_flavor = element.get('exec')
        if exec_flavor:
            self.exec_flavor = exec_flavor

        deprecated = element.get('deprecated', 'none')
        if deprecated != 'none':
            self.deprecated = Decimal(deprecated)

        if not is_attr_true(element, 'desktop', 'true'):
            self.desktop = False

        if self.has_no_error_variant or is_attr_true(element, 'no_error'):
            self.has_no_error_variant = True
        else:
            self.has_no_error_variant = False

        if alias:
            true_name = alias
        else:
            true_name = name

            self.has_hw_select_variant = exec_flavor == 'beginend' and name[0:6] == 'Vertex'

            # Only try to set the offset when a non-alias entry-point
            # is being processed.

            if name in static_data.offsets and static_data.offsets[name] <= static_data.MAX_OFFSETS:
                self.offset = static_data.offsets[name]
            elif name in static_data.offsets and static_data.offsets[name] > static_data.MAX_OFFSETS:
                self.offset = static_data.offsets[name]
                self.assign_offset = True
            else:
                if self.exec_flavor != "skip":
                    raise RuntimeError("Entry-point %s is missing offset in static_data.py. Add one at the bottom of the list." % (name))
                self.assign_offset = self.exec_flavor != "skip" or name in static_data.unused_functions

        if not self.name:
            self.name = true_name
        elif self.name != true_name:
            raise RuntimeError("Function true name redefined.  Was %s, now %s." % (self.name, true_name))


        # There are two possible cases.  The first time an entry-point
        # with data is seen, self.initialized will be 0.  On that
        # pass, we just fill in the data.  The next time an
        # entry-point with data is seen, self.initialized will be 1.
        # On that pass we have to make that the new values match the
        # valuse from the previous entry-point.

        parameters = []
        return_type = "void"
        for child in element:
            if child.tag == "return":
                return_type = child.get( "type", "void" )
            elif child.tag == "param":
                param = self.context.factory.create_parameter(child, self.context)
                parameters.append( param )


        if self.initialized:
            if self.return_type != return_type:
                raise RuntimeError( "Return type changed in %s.  Was %s, now %s." % (name, self.return_type, return_type))

            if len(parameters) != len(self.parameters):
                raise RuntimeError( "Parameter count mismatch in %s.  Was %d, now %d." % (name, len(self.parameters), len(parameters)))

            for j in range(0, len(parameters)):
                p1 = parameters[j]
                p2 = self.parameters[j]
                if not p1.compatible( p2 ):
                    raise RuntimeError( 'Parameter type mismatch in %s.  "%s" was "%s", now "%s".' % (name, p2.name, p2.type_expr.original_string, p1.type_expr.original_string))


        if true_name == name or not self.initialized:
            self.return_type = return_type
            self.parameters = parameters

            for param in self.parameters:
                if param.is_image():
                    self.images.append( param )

        if list(element):
            self.initialized = 1
            self.entry_point_parameters[name] = parameters
        else:
            self.entry_point_parameters[name] = []

        return

    def filter_entry_points(self, entry_point_list):
        """Filter out entry points not in entry_point_list."""
        if not self.initialized:
            raise RuntimeError('%s is not initialized yet' % self.name)

        entry_points = []
        for ent in self.entry_points:
            if ent not in entry_point_list:
                if ent in self.static_entry_points:
                    self.static_entry_points.remove(ent)
                self.entry_point_parameters.pop(ent)
            else:
                entry_points.append(ent)

        if not entry_points:
            raise RuntimeError('%s has no entry point after filtering' % self.name)

        self.entry_points = entry_points
        if self.name not in entry_points:
            # use the first remaining entry point
            self.name = entry_points[0]
            self.parameters = self.entry_point_parameters[entry_points[0]]

    def get_images(self):
        """Return potentially empty list of input images."""
        return self.images


    def parameterIterator(self, name = None):
        if name is not None:
            return iter(self.entry_point_parameters[name]);
        else:
            return iter(self.parameters);


    def get_parameter_string(self, entrypoint = None):
        if entrypoint:
            params = self.entry_point_parameters[ entrypoint ]
        else:
            params = self.parameters

        return create_parameter_string( params, 1 )

    def get_called_parameter_string(self):
        p_string = ""
        comma = ""

        for p in self.parameterIterator():
            if p.is_padding:
                continue
            p_string = p_string + comma + p.name
            comma = ", "

        return p_string


    def is_abi(self):
        return (self.offset >= 0 and not self.assign_offset)

    def is_static_entry_point(self, name):
        return name in self.static_entry_points

    def dispatch_name(self):
        if self.name in self.static_entry_points:
            return self.name
        else:
            return "_dispatch_stub_%u" % (self.offset)

    def static_name(self, name):
        if name in self.static_entry_points:
            return name
        else:
            return "_dispatch_stub_%u" % (self.offset)

class gl_item_factory(object):
    """Factory to create objects derived from gl_item."""

    def create_function(self, element, context):
        return gl_function(element, context)

    def create_type(self, element, context, category):
        return gl_type(element, context, category)

    def create_enum(self, element, context, category):
        return gl_enum(element, context, category)

    def create_parameter(self, element, context):
        return gl_parameter(element, context)

    def create_api(self):
        return gl_api(self)


class gl_api(object):
    def __init__(self, factory):
        self.functions_by_name = OrderedDict()
        self.enums_by_name = {}
        self.types_by_name = {}

        self.category_dict = {}
        self.categories = [{}, {}, {}, {}]

        self.factory = factory

        self.next_offset = 0

        typeexpr.create_initial_types()
        return

    def parse_file(self, file_name):
        doc = ET.parse( file_name )
        self.process_element(file_name, doc)


    def process_element(self, file_name, doc):
        element = doc.getroot()
        if element.tag == "OpenGLAPI":
            self.process_OpenGLAPI(file_name, element)
        return


    def process_OpenGLAPI(self, file_name, element):
        for child in element:
            if child.tag == "category":
                self.process_category( child )
            elif child.tag == "OpenGLAPI":
                self.process_OpenGLAPI( file_name, child )
            elif child.tag == '{http://www.w3.org/2001/XInclude}include':
                href = child.get('href')
                href = os.path.join(os.path.dirname(file_name), href)
                self.parse_file(href)

        return


    def process_category(self, cat):
        cat_name = cat.get( "name" )
        cat_number = cat.get( "number" )

        [cat_type, key] = classify_category(cat_name, cat_number)
        self.categories[cat_type][key] = [cat_name, cat_number]

        for child in cat:
            if child.tag == "function":
                func_name = real_function_name( child )

                temp_name = child.get( "name" )
                self.category_dict[ temp_name ] = [cat_name, cat_number]

                if func_name in self.functions_by_name:
                    func = self.functions_by_name[ func_name ]
                    func.process_element( child )
                else:
                    func = self.factory.create_function( child, self )
                    self.functions_by_name[ func_name ] = func

                if func.offset >= self.next_offset:
                    self.next_offset = func.offset + 1


            elif child.tag == "enum":
                enum = self.factory.create_enum( child, self, cat_name )
                self.enums_by_name[ enum.name ] = enum
            elif child.tag == "type":
                t = self.factory.create_type( child, self, cat_name )
                self.types_by_name[ "GL" + t.name ] = t

        return


    def functionIterateByCategory(self, cat = None):
        """Iterate over functions by category.

        If cat is None, all known functions are iterated in category
        order.  See classify_category for details of the ordering.
        Within a category, functions are sorted by name.  If cat is
        not None, then only functions in that category are iterated.
        """
        lists = [{}, {}, {}, {}]

        for func in self.functionIterateAll():
            [cat_name, cat_number] = self.category_dict[func.name]

            if (cat == None) or (cat == cat_name):
                [func_cat_type, key] = classify_category(cat_name, cat_number)

                if key not in lists[func_cat_type]:
                    lists[func_cat_type][key] = {}

                lists[func_cat_type][key][func.name] = func


        functions = []
        for func_cat_type in range(0,4):
            keys = sorted(lists[func_cat_type].keys())

            for key in keys:
                names = sorted(lists[func_cat_type][key].keys())

                for name in names:
                    functions.append(lists[func_cat_type][key][name])

        return iter(functions)


    def functionIterateByOffset(self):
        max_offset = -1
        for func in self.functions_by_name.values():
            if func.offset > max_offset:
                max_offset = func.offset


        temp = [None for i in range(0, max_offset + 1)]
        for func in self.functions_by_name.values():
            if func.offset != -1:
                temp[ func.offset ] = func


        list = []
        for i in range(0, max_offset + 1):
            if temp[i]:
                list.append(temp[i])

        return iter(list);


    def functionIterateAll(self):
        return self.functions_by_name.values()


    def enumIterateByName(self):
        keys = sorted(self.enums_by_name.keys())

        list = []
        for enum in keys:
            list.append( self.enums_by_name[ enum ] )

        return iter(list)


    def categoryIterate(self):
        """Iterate over categories.

        Iterate over all known categories in the order specified by
        classify_category.  Each iterated value is a tuple of the
        name and number (which may be None) of the category.
        """

        list = []
        for cat_type in range(0,4):
            keys = sorted(self.categories[cat_type].keys())

            for key in keys:
                list.append(self.categories[cat_type][key])

        return iter(list)


    def get_category_for_name( self, name ):
        if name in self.category_dict:
            return self.category_dict[name]
        else:
            return ["<unknown category>", None]


    def typeIterate(self):
        return self.types_by_name.values()


    def find_type( self, type_name ):
        if type_name in self.types_by_name:
            return self.types_by_name[ type_name ].type_expr
        else:
            print("Unable to find base type matching \"%s\"." % (type_name))
            return None



#!/usr/bin/env python3
# Copyright © 2020 - 2022 Collabora Ltd.
# Authors:
#   Tomeu Vizoso <tomeu.vizoso@collabora.com>
#   David Heidelberg <david.heidelberg@collabora.com>
#
# SPDX-License-Identifier: MIT
'''Shared functions between the scripts.'''

import os
import time
from typing import Optional


def get_gitlab_project(glab, name: str):
    """Finds a specified gitlab project for given user"""
    glab.auth()
    username = glab.user.username
    return glab.projects.get(f"{username}/mesa")


def read_token(token_arg: Optional[str]) -> str:
    """pick token from args or file"""
    if token_arg:
        return token_arg
    return (
        open(os.path.expanduser("~/.config/gitlab-token"), encoding="utf-8")
        .readline()
        .rstrip()
    )


def wait_for_pipeline(project, sha: str):
    """await until pipeline appears in Gitlab"""
    print("⏲ for the pipeline to appear..", end="")
    while True:
        pipelines = project.pipelines.list(sha=sha)
        if pipelines:
            print("", flush=True)
            return pipelines[0]
        print("", end=".", flush=True)
        time.sleep(1)



# Copyright (C) 2012 Intel Corporation
# Copyright (C) 2022 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import contextlib
import gl_XML
import license
import marshal_XML
import sys

header = """#include "glthread_marshal.h"
"""

current_indent = 0


def out(str):
    if str:
        print(' '*current_indent + str)
    else:
        print('')


@contextlib.contextmanager
def indent(delta=3):
    global current_indent
    current_indent += delta
    yield
    current_indent -= delta


class PrintCode(gl_XML.gl_print_base):
    def __init__(self):
        super(PrintCode, self).__init__()

        self.name = 'gl_marshal.py'
        self.license = license.bsd_license_template % (
            'Copyright (C) 2012 Intel Corporation', 'INTEL CORPORATION')

    def printRealHeader(self):
        print(header)

    def printRealFooter(self):
        pass

    def printBody(self, api):
        out('const _mesa_unmarshal_func _mesa_unmarshal_dispatch[NUM_DISPATCH_CMD] = {')
        with indent():
            for func in api.functionIterateAll():
                flavor = func.marshal_flavor()
                if flavor in ('skip', 'sync'):
                    continue
                out('[DISPATCH_CMD_{0}] = (_mesa_unmarshal_func)_mesa_unmarshal_{0},'.format(func.name))
        out('};')


def show_usage():
    print('Usage: %s [file_name]' % sys.argv[0])
    sys.exit(1)


if __name__ == '__main__':
    try:
        file_name = sys.argv[1]
    except Exception:
        show_usage()

    printer = PrintCode()

    api = gl_XML.parse_GL_API(file_name, marshal_XML.marshal_item_factory())
    printer.Print(api)



#!/usr/bin/env python3
#
# Copyright © 2021 Google LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import argparse
import io
import re
import socket
import time


class Connection:
    def __init__(self, host, port, verbose):
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.connect((host, port))
        self.s.setblocking(0)
        self.verbose = verbose

    def send_line(self, line):
        if self.verbose:
            print(f"IRC: sending {line}")
        self.s.sendall((line + '\n').encode())

    def wait(self, secs):
        for i in range(secs):
            if self.verbose:
                while True:
                    try:
                        data = self.s.recv(1024)
                    except io.BlockingIOError:
                        break
                    if data == "":
                        break
                    for line in data.decode().split('\n'):
                        print(f"IRC: received {line}")
            time.sleep(1)

    def quit(self):
        self.send_line("QUIT")
        self.s.shutdown(socket.SHUT_WR)
        self.s.close()


def read_flakes(results):
    flakes = []
    csv = re.compile("(.*),(.*),(.*)")
    for line in open(results, 'r').readlines():
        match = csv.match(line)
        if match.group(2) == "Flake":
            flakes.append(match.group(1))
    return flakes

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str,
                        help='IRC server hostname', required=True)
    parser.add_argument('--port', type=int,
                        help='IRC server port', required=True)
    parser.add_argument('--results', type=str,
                        help='results.csv file from deqp-runner or piglit-runner', required=True)
    parser.add_argument('--known-flakes', type=str,
                        help='*-flakes.txt file passed to deqp-runner or piglit-runner', required=True)
    parser.add_argument('--channel', type=str,
                        help='Known flakes report channel', required=True)
    parser.add_argument('--url', type=str,
                        help='$CI_JOB_URL', required=True)
    parser.add_argument('--runner', type=str,
                        help='$CI_RUNNER_DESCRIPTION', required=True)
    parser.add_argument('--branch', type=str,
                        help='optional branch name')
    parser.add_argument('--branch-title', type=str,
                        help='optional branch title')
    parser.add_argument('--job', type=str,
                        help='$CI_JOB_ID', required=True)
    parser.add_argument('--verbose', "-v", action="store_true",
                        help='log IRC interactions')
    args = parser.parse_args()

    flakes = read_flakes(args.results)
    if not flakes:
        exit(0)

    known_flakes = []
    for line in open(args.known_flakes).readlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        known_flakes.append(re.compile(line))

    irc = Connection(args.host, args.port, args.verbose)

    # The nick needs to be something unique so that multiple runners
    # connecting at the same time don't race for one nick and get blocked.
    # freenode has a 16-char limit on nicks (9 is the IETF standard, but
    # various servers extend that).  So, trim off the common prefixes of the
    # runner name, and append the job ID so that software runners with more
    # than one concurrent job (think swrast) don't collide.  For freedreno,
    # that gives us a nick as long as db410c-N-JJJJJJJJ, and it'll be a while
    # before we make it to 9-digit jobs (we're at 7 so far).
    nick = args.runner
    nick = nick.replace('mesa-', '')
    nick = nick.replace('google-freedreno-', '')
    nick += f'-{args.job}'
    irc.send_line(f"NICK {nick}")
    irc.send_line(f"USER {nick} unused unused: Gitlab CI Notifier")
    irc.wait(10)
    irc.send_line(f"JOIN {args.channel}")
    irc.wait(1)

    branchinfo = ""
    if args.branch:
        branchinfo = f" on branch {args.branch} ({args.branch_title})"
    irc.send_line(
        f"PRIVMSG {args.channel} :Flakes detected in job {args.url} on {args.runner}{branchinfo}:")

    for flake in flakes:
        status = "NEW "
        for known in known_flakes:
            if known.match(flake):
                status = ""
                break

        irc.send_line(f"PRIVMSG {args.channel} :{status}{flake}")

    irc.send_line(
        f"PRIVMSG {args.channel} :See {args.url}/artifacts/browse/results/")

    irc.quit()


if __name__ == '__main__':
    main()



#
# Copyright (C) 2020 Collabora, Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
import itertools
from bifrost_isa import parse_instructions, opname_to_c, expand_states
from mako.template import Template

instructions = parse_instructions(sys.argv[1], include_unused = True)

# Constructs a reserved mask for a derived to cull impossible encodings

def reserved_mask(derived):
    ((pos, width), opts) = derived
    reserved = [x is None for x in opts]
    mask = sum([(y << x) for x, y in enumerate(reserved)])
    return (pos, width, mask)

def reserved_masks(op):
    masks = [reserved_mask(m) for m in op[2].get("derived", [])]
    return [m for m in masks if m[2] != 0]

# To decode instructions, pattern match based on the rules:
#
# 1. Execution unit (FMA or ADD) must line up.
# 2. All exact bits must match.
# 3. No fields should be reserved in a legal encoding.
# 4. Tiebreaker: Longer exact masks (greater unsigned bitwise inverses) win.
#
# To implement, filter the execution unit and check for exact bits in
# descending order of exact mask length.  Check for reserved fields per
# candidate and succeed if it matches.
# found.

def decode_op(instructions, is_fma):
    # Filter out the desired execution unit
    options = [n for n in instructions.keys() if (n[0] == '*') == is_fma]

    # Sort by exact masks, descending
    MAX_MASK = (1 << (23 if is_fma else 20)) - 1
    options.sort(key = lambda n: (MAX_MASK ^ instructions[n][2]["exact"][0]))

    # Map to what we need to template
    mapped = [(opname_to_c(op), instructions[op][2]["exact"], reserved_masks(instructions[op])) for op in options]

    # Generate checks in order
    template = """void
bi_disasm_${unit}(FILE *fp, unsigned bits, struct bifrost_regs *srcs, struct bifrost_regs *next_regs, unsigned staging_register, unsigned branch_offset, struct bi_constants *consts, bool last)
{
    fputs("    ", fp);

% for (i, (name, (emask, ebits), derived)) in enumerate(options):
% if len(derived) > 0:
    ${"else " if i > 0 else ""}if (unlikely(((bits & ${hex(emask)}) == ${hex(ebits)})
% for (pos, width, reserved) in derived:
        && !(${hex(reserved)} & (1 << _BITS(bits, ${pos}, ${width})))
% endfor
    ))
% else:
    ${"else " if i > 0 else ""}if (unlikely(((bits & ${hex(emask)}) == ${hex(ebits)})))
% endif
        bi_disasm_${name}(fp, bits, srcs, next_regs, staging_register, branch_offset, consts, last);
% endfor
    else
        fprintf(fp, "INSTR_INVALID_ENC ${unit} %X", bits);

    fputs("\\n", fp);
}"""

    return Template(template).render(options = mapped, unit = "fma" if is_fma else "add")

# Decoding emits a series of function calls to e.g. `fma_fadd_v2f16`. We need to
# emit functions to disassemble a single decoded instruction in a particular
# state. Sync prototypes to avoid moves when calling.

disasm_op_template = Template("""static void
bi_disasm_${c_name}(FILE *fp, unsigned bits, struct bifrost_regs *srcs, struct bifrost_regs *next_regs, unsigned staging_register, unsigned branch_offset, struct bi_constants *consts, bool last)
{
    ${body.strip()}
}
""")

lut_template_only = Template("""    static const char *${field}[] = {
        ${", ".join(['"' + x + '"' for x in table])}
    };
""")

# Given a lookup table written logically, generate an accessor
lut_template = Template("""    static const char *${field}_table[] = {
        ${", ".join(['"' + x + '"' for x in table])}
    };

    const char *${field} = ${field}_table[_BITS(bits, ${pos}, ${width})];
""")

# Helpers for decoding follow. pretty_mods applies dot syntax

def pretty_mods(opts, default):
    return [('.' + (opt or 'reserved') if opt != default else '') for opt in opts]

# Recursively searches for the set of free variables required by an expression

def find_context_keys_expr(expr):
    if isinstance(expr, list):
        return set.union(*[find_context_keys_expr(x) for x in expr[1:]])
    elif expr[0] == '#':
        return set()
    else:
        return set([expr])

def find_context_keys(desc, test):
    keys = set()

    if len(test) > 0:
        keys |= find_context_keys_expr(test)

    for i, (_, vals) in enumerate(desc.get('derived', [])):
        for j, val in enumerate(vals):
            if val is not None:
                keys |= find_context_keys_expr(val)

    return keys
 
# Compiles a logic expression to Python expression, ctx -> { T, F }

EVALUATORS = {
        'and': ' and ',
        'or': ' or ',
        'eq': ' == ',
        'neq': ' != ',
}

def compile_derived_inner(expr, keys):
    if expr == []:
        return 'True'
    elif expr is None or expr[0] == 'alias':
        return 'False'
    elif isinstance(expr, list):
        args = [compile_derived_inner(arg, keys) for arg in expr[1:]]
        return '(' + EVALUATORS[expr[0]].join(args) + ')'
    elif expr[0] == '#':
        return "'{}'".format(expr[1:])
    elif expr == 'ordering':
        return expr
    else:
        return "ctx[{}]".format(keys.index(expr))

def compile_derived(expr, keys):
    return eval('lambda ctx, ordering: ' + compile_derived_inner(expr, keys))

# Generate all possible combinations of values and evaluate the derived values
# by bruteforce evaluation to generate a forward mapping (values -> deriveds)

def evaluate_forward_derived(vals, ctx, ordering):
    for j, expr in enumerate(vals):
        if expr(ctx, ordering):
            return j

    return None

def evaluate_forward(keys, derivf, testf, ctx, ordering):
    if not testf(ctx, ordering):
        return None

    deriv = []

    for vals in derivf:
        evaled = evaluate_forward_derived(vals, ctx, ordering) 

        if evaled is None:
            return None

        deriv.append(evaled)

    return deriv

def evaluate_forwards(keys, derivf, testf, mod_vals, ordered):
    orderings = ["lt", "gt"] if ordered else [None]
    return [[evaluate_forward(keys, derivf, testf, i, order) for i in itertools.product(*mod_vals)] for order in orderings]

# Invert the forward mapping (values -> deriveds) of finite sets to produce a
# backwards mapping (deriveds -> values), suitable for disassembly. This is
# possible since the encoding is unambiguous, so this mapping is a bijection
# (after reserved/impossible encodings)

def invert_lut(value_size, forward, derived, mod_map, keys, mod_vals):
    backwards = [None] * (1 << value_size)
    for (i, deriveds), ctx in zip(enumerate(forward), itertools.product(*mod_vals)):
        # Skip reserved
        if deriveds == None:
            continue

        shift = 0
        param = 0
        for j, ((x, width), y) in enumerate(derived):
            param += (deriveds[j] << shift)
            shift += width

        assert(param not in backwards)
        backwards[param] = ctx

    return backwards

# Compute the value of all indirectly specified modifiers by using the
# backwards mapping (deriveds -> values) as a run-time lookup table.

def build_lut(mnemonic, desc, test):
    # Construct the system
    facts = []

    mod_map = {}

    for ((name, pos, width), default, values) in desc.get('modifiers', []):
        mod_map[name] = (width, values, pos, default)

    derived = desc.get('derived', [])

    # Find the keys and impose an order
    key_set = find_context_keys(desc, test)
    ordered = 'ordering' in key_set
    key_set.discard('ordering')
    keys = sorted(list(key_set))

    # Evaluate the deriveds for every possible state, forming a (state -> deriveds) map
    testf = compile_derived(test, keys)
    derivf = [[compile_derived(expr, keys) for expr in v] for (_, v) in derived]
    mod_vals = [mod_map[k][1] for k in keys]
    forward = evaluate_forwards(keys, derivf, testf, mod_vals, ordered)

    # Now invert that map to get a (deriveds -> state) map
    value_size = sum([width for ((x, width), y) in derived])
    backwards = [invert_lut(value_size, f, derived, mod_map, keys, mod_vals) for f in forward]

    # From that map, we can generate LUTs
    output = ""

    if ordered:
        output += "bool ordering = (_BITS(bits, {}, 3) > _BITS(bits, {}, 3));\n".format(desc["srcs"][0][0], desc["srcs"][1][0])

    for j, key in enumerate(keys):
        # Only generate tables for indirect specifiers
        if mod_map[key][2] is not None:
            continue

        idx_parts = []
        shift = 0

        for ((pos, width), _) in derived:
            idx_parts.append("(_BITS(bits, {}, {}) << {})".format(pos, width, shift))
            shift += width

        built_idx = (" | ".join(idx_parts)) if len(idx_parts) > 0 else "0"

        default = mod_map[key][3]

        if ordered:
            for i, order in enumerate(backwards):
                options = [ctx[j] if ctx is not None and ctx[j] is not None else "reserved" for ctx in order]
                output += lut_template_only.render(field = key + "_" + str(i), table = pretty_mods(options, default))

            output += "    const char *{} = ordering ? {}_1[{}] : {}_0[{}];\n".format(key, key, built_idx, key, built_idx)
        else:
            options = [ctx[j] if ctx is not None and ctx[j] is not None else "reserved" for ctx in backwards[0]]
            output += lut_template_only.render(field = key + "_table", table = pretty_mods(options, default))
            output += "    const char *{} = {}_table[{}];\n".format(key, key, built_idx)

    return output

def disasm_mod(mod, skip_mods):
    if mod[0][0] in skip_mods:
        return ''
    else:
        return '    fputs({}, fp);\n'.format(mod[0][0])

def disasm_op(name, op):
    (mnemonic, test, desc) = op
    is_fma = mnemonic[0] == '*'

    # Modifiers may be either direct (pos is not None) or indirect (pos is
    # None). If direct, we just do the bit lookup. If indirect, we use a LUT.

    body = ""
    skip_mods = []

    body += build_lut(mnemonic, desc, test)

    for ((mod, pos, width), default, opts) in desc.get('modifiers', []):
        if pos is not None:
            body += lut_template.render(field = mod, table = pretty_mods(opts, default), pos = pos, width = width) + "\n"

    # Mnemonic, followed by modifiers
    body += '    fputs("{}", fp);\n'.format(mnemonic)

    srcs = desc.get('srcs', [])

    for mod in desc.get('modifiers', []):
        # Skip per-source until next block
        if mod[0][0][-1] in "0123" and int(mod[0][0][-1]) < len(srcs):
            continue

        body += disasm_mod(mod, skip_mods)

    body += '    fputs(" ", fp);\n'
    body += '    bi_disasm_dest_{}(fp, next_regs, last);\n'.format('fma' if is_fma else 'add')

    # Next up, each source. Source modifiers are inserterd here

    for i, (pos, mask) in enumerate(srcs):
        body += '    fputs(", ", fp);\n'
        body += '    dump_src(fp, _BITS(bits, {}, 3), *srcs, branch_offset, consts, {});\n'.format(pos, "true" if is_fma else "false")

        # Error check if needed
        if (mask != 0xFF):
            body += '    if (!({} & (1 << _BITS(bits, {}, 3)))) fputs("(INVALID)", fp);\n'.format(hex(mask), pos, 3)

        # Print modifiers suffixed with this src number (e.g. abs0 for src0)
        for mod in desc.get('modifiers', []):
            if mod[0][0][-1] == str(i):
                body += disasm_mod(mod, skip_mods)

    # And each immediate
    for (imm, pos, width) in desc.get('immediates', []):
        body += '    fprintf(fp, ", {}:%u", _BITS(bits, {}, {}));\n'.format(imm, pos, width)

    # Attach a staging register if one is used
    if desc.get('staging'):
        body += '    fprintf(fp, ", @r%u", staging_register);\n'

    return disasm_op_template.render(c_name = opname_to_c(name), body = body)

print('#include "util/macros.h"')
print('#include "disassemble.h"')

states = expand_states(instructions)
print('#define _BITS(bits, pos, width) (((bits) >> (pos)) & ((1 << (width)) - 1))')

for st in states:
    print(disasm_op(st, states[st]))

print(decode_op(states, True))
print(decode_op(states, False))



#
# Copyright 2017 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# on the rights to use, copy, modify, merge, publish, distribute, sub
# license, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# THE AUTHOR(S) AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
#
"""
Script that generates the mapping from Gallium PIPE_FORMAT_xxx to GFX10_FORMAT_xxx enums.
"""

import json
import mako.template
import os
import re
import sys

AMD_REGISTERS = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), "../registers"))
UTIL_FORMAT = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), "../../util/format"))
sys.path.extend([AMD_REGISTERS, UTIL_FORMAT])

from regdb import Object, RegisterDatabase
from u_format_parse import *

# ----------------------------------------------------------------------------
# Hard-coded mappings

def hardcoded_format(hw_enum):
    return Object(img_format=hw_enum, flags=[])

HARDCODED = {
    'PIPE_FORMAT_Z32_FLOAT_S8X24_UINT': hardcoded_format('X24_8_32_FLOAT'),
    'PIPE_FORMAT_Z24_UNORM_S8_UINT': hardcoded_format('8_24_UNORM'),
    'PIPE_FORMAT_S8_UINT_Z24_UNORM': hardcoded_format('24_8_UNORM'),
    'PIPE_FORMAT_Z32_UNORM': None,
    'PIPE_FORMAT_Z16_UNORM_S8_UINT': None,

    'PIPE_FORMAT_R9G9B9E5_FLOAT': hardcoded_format('5_9_9_9_FLOAT'),
    'PIPE_FORMAT_R11G11B10_FLOAT': hardcoded_format('10_11_11_FLOAT'), # NOTE: full set of int/unorm/etc. exists

    'PIPE_FORMAT_R8G8_B8G8_UNORM': hardcoded_format('GB_GR_UNORM'),
    'PIPE_FORMAT_G8R8_B8R8_UNORM': hardcoded_format('GB_GR_UNORM'),

    'PIPE_FORMAT_R8G8_R8B8_UNORM': hardcoded_format('BG_RG_UNORM'),
    'PIPE_FORMAT_G8R8_G8B8_UNORM': hardcoded_format('BG_RG_UNORM'),

    # These mixed channel types are not supported natively
    'PIPE_FORMAT_R8SG8SB8UX8U_NORM': None,
    'PIPE_FORMAT_R10SG10SB10SA2U_NORM': None,
    'PIPE_FORMAT_R5SG5SB6U_NORM': None,

    # Only R8G8_SRGB is supported, not L8A8_SRGB
    'PIPE_FORMAT_L8A8_SRGB': None,

    # S3TC
    'PIPE_FORMAT_DXT1_RGB': hardcoded_format('BC1_UNORM'),
    'PIPE_FORMAT_DXT1_RGBA': hardcoded_format('BC1_UNORM'),
    'PIPE_FORMAT_DXT1_SRGB': hardcoded_format('BC1_SRGB'),
    'PIPE_FORMAT_DXT1_SRGBA': hardcoded_format('BC1_SRGB'),
    'PIPE_FORMAT_DXT3_RGBA': hardcoded_format('BC2_UNORM'),
    'PIPE_FORMAT_DXT3_SRGBA': hardcoded_format('BC2_SRGB'),
    'PIPE_FORMAT_DXT5_RGBA': hardcoded_format('BC3_UNORM'),
    'PIPE_FORMAT_DXT5_SRGBA': hardcoded_format('BC3_SRGB'),

    # RGTC
    'PIPE_FORMAT_RGTC1_UNORM': hardcoded_format('BC4_UNORM'),
    'PIPE_FORMAT_RGTC1_SNORM': hardcoded_format('BC4_SNORM'),
    'PIPE_FORMAT_RGTC2_UNORM': hardcoded_format('BC5_UNORM'),
    'PIPE_FORMAT_RGTC2_SNORM': hardcoded_format('BC5_SNORM'),
    'PIPE_FORMAT_LATC1_UNORM': hardcoded_format('BC4_UNORM'),
    'PIPE_FORMAT_LATC1_SNORM': hardcoded_format('BC4_SNORM'),
    'PIPE_FORMAT_LATC2_UNORM': hardcoded_format('BC5_UNORM'),
    'PIPE_FORMAT_LATC2_SNORM': hardcoded_format('BC5_SNORM'),

    # BPTC
    'PIPE_FORMAT_BPTC_RGB_UFLOAT': hardcoded_format('BC6_UFLOAT'),
    'PIPE_FORMAT_BPTC_RGB_FLOAT': hardcoded_format('BC6_SFLOAT'),

    'PIPE_FORMAT_BPTC_RGBA_UNORM': hardcoded_format('BC7_UNORM'),
    'PIPE_FORMAT_BPTC_SRGBA': hardcoded_format('BC7_SRGB'),

    'PIPE_FORMAT_R64_UINT': hardcoded_format('32_32_UINT'),
    'PIPE_FORMAT_R64_SINT': hardcoded_format('32_32_SINT'),
}


# ----------------------------------------------------------------------------
# Main script

header_template = mako.template.Template("""\
% if header:
// DO NOT EDIT -- AUTOMATICALLY GENERATED

#include "gfx10_format_table.h"
#include "amdgfxregs.h"

% endif

#define FMT(_img_format, ...) \
   { .img_format = V_008F0C_${gfx.upper()}_FORMAT_##_img_format, \
     ##__VA_ARGS__ }

const struct gfx10_format ${gfx}_format_table[PIPE_FORMAT_COUNT] = {
% for pipe_format, args in formats:
 % if args is not None:
  [${pipe_format}] = FMT(${args}),
 % else:
/* ${pipe_format} is not supported */
 % endif
% endfor

#undef FMT
};
""")

class Gfx10Format(object):
    RE_plain_channel = re.compile(r'X?([0-9]+)')

    def __init__(self, enum_entry):
        self.img_format = enum_entry.name[13:]
        self.flags = getattr(enum_entry, 'flags', [])

        code = self.img_format.split('_')

        self.plain_chan_sizes = []
        for i, chan_code in enumerate(code):
            m = self.RE_plain_channel.match(chan_code)
            if m is None:
                break
            self.plain_chan_sizes.append(int(m.group(1)))
        # Keep the bit sizes in little-endian order
        self.plain_chan_sizes.reverse()

        self.code = code[i:]


class Gfx10FormatMapping(object):
    def __init__(self, pipe_formats, gfx10_formats):
        self.pipe_formats = pipe_formats
        self.gfx10_formats = gfx10_formats

        self.plain_gfx10_formats = dict(
            (tuple(['_'.join(fmt.code)] + fmt.plain_chan_sizes), fmt)
            for fmt in gfx10_formats if fmt.plain_chan_sizes
        )

    def map(self, fmt):
        if fmt.layout == PLAIN:
            chan_type = set([chan.type for chan in fmt.le_channels if chan.type != VOID])
            chan_norm = set([chan.norm for chan in fmt.le_channels if chan.type != VOID])
            chan_pure = set([chan.pure for chan in fmt.le_channels if chan.type != VOID])
            if len(chan_type) > 1 or len(chan_norm) > 1 or len(chan_pure) > 1:
                print(('Format {fmt.name} has inconsistent channel types: ' +
                        '{chan_type} {chan_norm} {chan_pure}')
                      .format(**locals()),
                      file=sys.stderr)
                return None

            chan_type = chan_type.pop()
            chan_norm = chan_norm.pop()
            chan_pure = chan_pure.pop()
            chan_sizes = [chan.size for chan in fmt.le_channels if chan.size != 0]

            extra_flags = []

            if fmt.colorspace == SRGB:
                assert chan_type == UNSIGNED and chan_norm
                num_format = 'SRGB'
            else:
                if chan_type == UNSIGNED:
                    if chan_pure:
                        num_format = 'UINT'
                    elif chan_sizes[0] == 32:
                        # Shader-based work-around for 32-bit non-pure-integer
                        num_format = 'UINT'
                        extra_flags.append('buffers_only')
                    elif chan_norm:
                        num_format = 'UNORM'
                    else:
                        num_format = 'USCALED'
                        extra_flags.append('buffers_only')
                elif chan_type == SIGNED:
                    if chan_pure:
                        num_format = 'SINT'
                    elif chan_sizes[0] == 32:
                        # Shader-based work-around for 32-bit non-pure-integer
                        num_format = 'SINT'
                        extra_flags.append('buffers_only')
                    elif chan_norm:
                        num_format = 'SNORM'
                    else:
                        num_format = 'SSCALED'
                        extra_flags.append('buffers_only')
                elif chan_type == FLOAT:
                    num_format = 'FLOAT'

                    if chan_sizes[0] == 64:
                        # Shader-based work-around for doubles
                        if len(chan_sizes) % 2 == 1:
                            # 1 or 3 loads for 1 or 3 double channels
                            chan_sizes = [32, 32]
                        else:
                            # 1 or 2 loads for 2 or 4 double channels
                            chan_sizes = [32, 32, 32, 32]
                        extra_flags.append('buffers_only')
                else:
                    # Shader-based work-around
                    assert chan_type == FIXED
                    assert chan_sizes[0] == 32
                    num_format = 'SINT'
                    extra_flags.append('buffers_only')

            # These are not supported as render targets, so we don't support
            # them as images either.
            if (len(chan_sizes) == 3 and chan_sizes[0] in (8, 16, 32) and
                chan_sizes[0] == chan_sizes[1]):
                extra_flags.append('buffers_only')
                if chan_sizes[0] in (8, 16):
                    # Shader-based work-around: one load per channel
                    chan_sizes = [chan_sizes[0]]

            # Don't expose SRGB buffer formats
            if 'buffers_only' in extra_flags and fmt.colorspace == SRGB:
                return None

            # Don't support 4_4 because it's not supported as render targets
            # and it's useless in other cases.
            if len(chan_sizes) == 2 and chan_sizes[0] == 4:
                return None

            key = tuple([num_format] + chan_sizes)
            if key not in self.plain_gfx10_formats:
                return None

            gfx10_fmt = self.plain_gfx10_formats[key]
            return Object(
                img_format=gfx10_fmt.img_format,
                flags=gfx10_fmt.flags + extra_flags,
            )

        return None

def pipe_formats_to_formats(pipe_formats, mapping):
    formats = []
    for fmt in pipe_formats:
        if fmt.name in HARDCODED:
            obj = HARDCODED[fmt.name]
        else:
            obj = mapping.map(fmt)

        if obj is not None:
            args = obj.img_format
            if 'buffers_only' in obj.flags:
                args += ', .buffers_only = 1'
        else:
            args = None
        formats.append((fmt.name, args))

    return formats

if __name__ == '__main__':
    pipe_formats = parse(sys.argv[1])

    # gfx10
    with open(sys.argv[2], 'r') as filp:
        db = RegisterDatabase.from_json(json.load(filp))

    gfx10_formats = [Gfx10Format(entry) for entry in db.enum('GFX10_FORMAT').entries]
    mapping = Gfx10FormatMapping(pipe_formats, gfx10_formats)
    formats = pipe_formats_to_formats(pipe_formats, mapping)
    print(header_template.render(header=True, gfx='gfx10', formats=formats))

    # gfx11
    with open(sys.argv[3], 'r') as filp:
        db = RegisterDatabase.from_json(json.load(filp))

    gfx11_formats = [Gfx10Format(entry) for entry in db.enum('GFX11_FORMAT').entries]
    mapping = Gfx10FormatMapping(pipe_formats, gfx11_formats)
    formats = pipe_formats_to_formats(pipe_formats, mapping)
    print(header_template.render(header=False, gfx='gfx11', formats=formats))



#
# Copyright (C) 2020 Google, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

from mako.template import Template
from collections import namedtuple
from enum import IntEnum
import os

TRACEPOINTS = {}
TRACEPOINTS_TOGGLES = {}

class Tracepoint(object):
    """Class that represents all the information about a tracepoint
    """
    def __init__(self, name, args=[], toggle_name=None,
                 tp_struct=None, tp_print=None, tp_perfetto=None,
                 end_of_pipe=False):
        """Parameters:

        - name: the tracepoint name, a tracepoint function with the given
          name (prefixed by 'trace_') will be generated with the specied
          args (following a u_trace ptr).  Calling this tracepoint will
          emit a trace, if tracing is enabled.
        - args: the tracepoint func args, an array of TracepointArg
        - tp_print: (optional) array of format string followed by expressions
        - tp_perfetto: (optional) driver provided callback which can generate
          perfetto events
        """
        assert isinstance(name, str)
        assert isinstance(args, list)
        assert name not in TRACEPOINTS

        self.name = name
        self.args = args
        if tp_struct is None:
           tp_struct = args
        self.tp_struct = tp_struct
        self.tp_print = tp_print
        self.tp_perfetto = tp_perfetto
        self.end_of_pipe = end_of_pipe
        self.toggle_name = toggle_name

        TRACEPOINTS[name] = self
        if toggle_name is not None and toggle_name not in TRACEPOINTS_TOGGLES:
            TRACEPOINTS_TOGGLES[toggle_name] = len(TRACEPOINTS_TOGGLES)

    def can_generate_print(self):
        return self.args is not None and len(self.args) > 0

    def enabled_expr(self, trace_toggle_name):
        if trace_toggle_name is None:
            return "true"
        assert self.toggle_name is not None
        return "({0} & {1}_{2})".format(trace_toggle_name,
                                        trace_toggle_name.upper(),
                                        self.toggle_name.upper())

class TracepointArgStruct():
    """Represents struct that is being passed as an argument
    """
    def __init__(self, type, var):
        """Parameters:

        - type: argument's C type.
        - var: name of the argument
        """
        assert isinstance(type, str)
        assert isinstance(var, str)

        self.type = type
        self.var = var

class TracepointArg(object):
    """Class that represents either an argument being passed or a field in a struct
    """
    def __init__(self, type, var, c_format, name=None, to_prim_type=None):
        """Parameters:

        - type: argument's C type.
        - var: either an argument name or a field in the struct
        - c_format: printf format to print the value.
        - name: (optional) name that will be used in intermidiate structs and will
          be displayed in output or perfetto, otherwise var will be used.
        - to_prim_type: (optional) C function to convert from arg's type to a type
          compatible with c_format.
        """
        assert isinstance(type, str)
        assert isinstance(var, str)
        assert isinstance(c_format, str)

        self.type = type
        self.var = var
        self.c_format = c_format
        if name is None:
           name = var
        self.name = name
        self.to_prim_type = to_prim_type


HEADERS = []

class HeaderScope(IntEnum):
    HEADER = (1 << 0)
    SOURCE = (1 << 1)
    PERFETTO = (1 << 2)

class Header(object):
    """Class that represents a header file dependency of generated tracepoints
    """
    def __init__(self, hdr, scope=HeaderScope.HEADER):
        """Parameters:

        - hdr: the required header path
        """
        assert isinstance(hdr, str)
        self.hdr = hdr
        self.scope = scope

        HEADERS.append(self)


FORWARD_DECLS = []

class ForwardDecl(object):
   """Class that represents a forward declaration
   """
   def __init__(self, decl):
        assert isinstance(decl, str)
        self.decl = decl

        FORWARD_DECLS.append(self)


hdr_template = """\
/* Copyright (C) 2020 Google, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

<% guard_name = '_' + hdrname + '_H' %>
#ifndef ${guard_name}
#define ${guard_name}

% for header in HEADERS:
#include "${header.hdr}"
% endfor

#include "util/perf/u_trace.h"

#ifdef __cplusplus
extern "C" {
#endif

% for declaration in FORWARD_DECLS:
${declaration.decl};
% endfor

% if trace_toggle_name is not None:
enum ${trace_toggle_name.lower()} {
%    for toggle_name, config_id in TRACEPOINTS_TOGGLES.items():
   ${trace_toggle_name.upper()}_${toggle_name.upper()} = 1ull << ${config_id},
%    endfor
};

extern uint64_t ${trace_toggle_name};

void ${trace_toggle_name}_config_variable(void);
% endif

% for trace_name, trace in TRACEPOINTS.items():

/*
 * ${trace_name}
 */
struct trace_${trace_name} {
%    for arg in trace.tp_struct:
   ${arg.type} ${arg.name};
%    endfor
%    if len(trace.args) == 0:
#ifdef __cplusplus
   /* avoid warnings about empty struct size mis-match in C vs C++..
    * the size mis-match is harmless because (a) nothing will deref
    * the empty struct, and (b) the code that cares about allocating
    * sizeof(struct trace_${trace_name}) (and wants this to be zero
    * if there is no payload) is C
    */
   uint8_t dummy;
#endif
%    endif
};
%    if trace.tp_perfetto is not None:
#ifdef HAVE_PERFETTO
void ${trace.tp_perfetto}(
   ${ctx_param},
   uint64_t ts_ns,
   const void *flush_data,
   const struct trace_${trace_name} *payload);
#endif
%    endif
void __trace_${trace_name}(
       struct u_trace *ut
%    if need_cs_param:
     , void *cs
%    endif
%    for arg in trace.args:
     , ${arg.type} ${arg.var}
%    endfor
);
static ALWAYS_INLINE void trace_${trace_name}(
     struct u_trace *ut
%    if need_cs_param:
   , void *cs
%    endif
%    for arg in trace.args:
   , ${arg.type} ${arg.var}
%    endfor
) {
   if (!unlikely(u_trace_instrument() &&
                 ${trace.enabled_expr(trace_toggle_name)}))
      return;
   __trace_${trace_name}(
        ut
%    if need_cs_param:
      , cs
%    endif
%    for arg in trace.args:
      , ${arg.var}
%    endfor
   );
}
% endfor

#ifdef __cplusplus
}
#endif

#endif /* ${guard_name} */
"""

src_template = """\
/* Copyright (C) 2020 Google, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include "${hdr}"

% for header in HEADERS:
#include "${header.hdr}"
% endfor

#define __NEEDS_TRACE_PRIV
#include "util/u_debug.h"
#include "util/perf/u_trace_priv.h"

% if trace_toggle_name is not None:
static const struct debug_control config_control[] = {
%    for toggle_name in TRACEPOINTS_TOGGLES.keys():
   { "${toggle_name}", ${trace_toggle_name.upper()}_${toggle_name.upper()}, },
%    endfor
   { NULL, 0, },
};
uint64_t ${trace_toggle_name} = 0;

static void
${trace_toggle_name}_variable_once(void)
{
   uint64_t default_value = 0
%    for name in trace_toggle_defaults:
     | ${trace_toggle_name.upper()}_${name.upper()}
%    endfor
     ;

   ${trace_toggle_name} =
      parse_enable_string(getenv("${trace_toggle_name.upper()}"),
                          default_value,
                          config_control);
}

void
${trace_toggle_name}_config_variable(void)
{
   static once_flag process_${trace_toggle_name}_variable_flag = ONCE_FLAG_INIT;

   call_once(&process_${trace_toggle_name}_variable_flag,
             ${trace_toggle_name}_variable_once);
}
% endif

% for trace_name, trace in TRACEPOINTS.items():
/*
 * ${trace_name}
 */
 % if trace.can_generate_print():
static void __print_${trace_name}(FILE *out, const void *arg) {
   const struct trace_${trace_name} *__entry =
      (const struct trace_${trace_name} *)arg;
  % if trace.tp_print is not None:
   fprintf(out, "${trace.tp_print[0]}\\n"
   % for arg in trace.tp_print[1:]:
           , ${arg}
   % endfor
  % else:
   fprintf(out, ""
   % for arg in trace.tp_struct:
      "${arg.name}=${arg.c_format}, "
   % endfor
         "\\n"
   % for arg in trace.tp_struct:
    % if arg.to_prim_type:
   ,${arg.to_prim_type.format('__entry->' + arg.name)}
    % else:
   ,__entry->${arg.name}
    % endif
   % endfor
  % endif
   );
}

static void __print_json_${trace_name}(FILE *out, const void *arg) {
   const struct trace_${trace_name} *__entry =
      (const struct trace_${trace_name} *)arg;
  % if trace.tp_print is not None:
   fprintf(out, "\\"unstructured\\": \\"${trace.tp_print[0]}\\""
   % for arg in trace.tp_print[1:]:
           , ${arg}
   % endfor
  % else:
   fprintf(out, ""
   % for arg in trace.tp_struct:
      "\\"${arg.name}\\": \\"${arg.c_format}\\""
      % if arg != trace.tp_struct[-1]:
         ", "
      % endif
   % endfor
   % for arg in trace.tp_struct:
    % if arg.to_prim_type:
   ,${arg.to_prim_type.format('__entry->' + arg.name)}
    % else:
   ,__entry->${arg.name}
    % endif
   % endfor
  % endif
   );
}

 % else:
#define __print_${trace_name} NULL
#define __print_json_${trace_name} NULL
 % endif
static const struct u_tracepoint __tp_${trace_name} = {
    ALIGN_POT(sizeof(struct trace_${trace_name}), 8),   /* keep size 64b aligned */
    "${trace_name}",
    ${"true" if trace.end_of_pipe else "false"},
    __print_${trace_name},
    __print_json_${trace_name},
 % if trace.tp_perfetto is not None:
#ifdef HAVE_PERFETTO
    (void (*)(void *pctx, uint64_t, const void *, const void *))${trace.tp_perfetto},
#endif
 % endif
};
void __trace_${trace_name}(
     struct u_trace *ut
 % if need_cs_param:
   , void *cs
 % endif
 % for arg in trace.args:
   , ${arg.type} ${arg.var}
 % endfor
) {
   struct trace_${trace_name} *__entry =
      (struct trace_${trace_name} *)u_trace_append(ut, ${cs_param_value + ","} &__tp_${trace_name});
 % if len(trace.tp_struct) == 0:
   (void)__entry;
 % endif
 % for arg in trace.tp_struct:
   __entry->${arg.name} = ${arg.var};
 % endfor
}

% endfor
"""

def utrace_generate(cpath, hpath, ctx_param, need_cs_param=True,
                    trace_toggle_name=None, trace_toggle_defaults=[]):
    """Parameters:

    - cpath: c file to generate.
    - hpath: h file to generate.
    - ctx_param: type of the first parameter to the perfetto vfuncs.
    - need_cs_param: whether tracepoint functions need an additional cs
      parameter.
    - trace_toggle_name: (optional) name of the environment variable
      enabling/disabling tracepoints.
    - trace_toggle_defaults: (optional) list of tracepoints enabled by default.
    """
    cs_param_value = 'NULL'
    if need_cs_param:
        cs_param_value = 'cs'
    if cpath is not None:
        hdr = os.path.basename(cpath).rsplit('.', 1)[0] + '.h'
        with open(cpath, 'w') as f:
            f.write(Template(src_template).render(
                hdr=hdr,
                ctx_param=ctx_param,
                need_cs_param=need_cs_param,
                cs_param_value=cs_param_value,
                trace_toggle_name=trace_toggle_name,
                trace_toggle_defaults=trace_toggle_defaults,
                HEADERS=[h for h in HEADERS if h.scope & HeaderScope.SOURCE],
                TRACEPOINTS=TRACEPOINTS,
                TRACEPOINTS_TOGGLES=TRACEPOINTS_TOGGLES))

    if hpath is not None:
        hdr = os.path.basename(hpath)
        with open(hpath, 'w') as f:
            f.write(Template(hdr_template).render(
                hdrname=hdr.rstrip('.h').upper(),
                ctx_param=ctx_param,
                need_cs_param=need_cs_param,
                trace_toggle_name=trace_toggle_name,
                HEADERS=[h for h in HEADERS if h.scope & HeaderScope.HEADER],
                FORWARD_DECLS=FORWARD_DECLS,
                TRACEPOINTS=TRACEPOINTS,
                TRACEPOINTS_TOGGLES=TRACEPOINTS_TOGGLES))


perfetto_utils_hdr_template = """\
/*
 * Copyright © 2021 Igalia S.L.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

<% guard_name = '_' + hdrname + '_H' %>
#ifndef ${guard_name}
#define ${guard_name}

#include <perfetto.h>

% for header in HEADERS:
#include "${header.hdr}"
% endfor

% for trace_name, trace in TRACEPOINTS.items():
static void UNUSED
trace_payload_as_extra_${trace_name}(perfetto::protos::pbzero::GpuRenderStageEvent *event,
                                     const struct trace_${trace_name} *payload)
{
 % if all([trace.tp_perfetto, trace.tp_struct]) and len(trace.tp_struct) > 0:
   char buf[128];

  % for arg in trace.tp_struct:
   {
      auto data = event->add_extra_data();
      data->set_name("${arg.name}");

   % if arg.to_prim_type:
      sprintf(buf, "${arg.c_format}", ${arg.to_prim_type.format('payload->' + arg.name)});
   % else:
      sprintf(buf, "${arg.c_format}", payload->${arg.name});
   % endif

      data->set_value(buf);
   }
  % endfor

 % endif
}
% endfor

#endif /* ${guard_name} */
"""

def utrace_generate_perfetto_utils(hpath):
    if hpath is not None:
        hdr = os.path.basename(hpath)
        with open(hpath, 'wb') as f:
            f.write(Template(perfetto_utils_hdr_template, output_encoding='utf-8').render(
                hdrname=hdr.rstrip('.h').upper(),
                HEADERS=[h for h in HEADERS if h.scope & HeaderScope.PERFETTO],
                TRACEPOINTS=TRACEPOINTS))



#!/usr/bin/python3

import xml.parsers.expat
import sys
import os

class Error(Exception):
	def __init__(self, message):
		self.message = message

class Enum(object):
	def __init__(self, name):
		self.name = name
		self.values = []

	def dump(self):
		prev = 0
		use_hex = False
		for (name, value) in self.values:
			if value > 0x1000:
				use_hex = True

		print("enum %s {" % self.name)
		for (name, value) in self.values:
			if use_hex:
				print("\t%s = 0x%08x," % (name, value))
			else:
				print("\t%s = %d," % (name, value))
		print("};\n")

	def dump_pack_struct(self):
		pass

class Field(object):
	def __init__(self, name, low, high, shr, type, parser):
		self.name = name
		self.low = low
		self.high = high
		self.shr = shr
		self.type = type

		builtin_types = [ None, "a3xx_regid", "boolean", "uint", "hex", "int", "fixed", "ufixed", "float", "address", "waddress" ]

		if low < 0 or low > 31:
			raise parser.error("low attribute out of range: %d" % low)
		if high < 0 or high > 31:
			raise parser.error("high attribute out of range: %d" % high)
		if high < low:
			raise parser.error("low is greater than high: low=%d, high=%d" % (low, high))
		if self.type == "boolean" and not low == high:
			raise parser.error("booleans should be 1 bit fields");
		elif self.type == "float" and not (high - low == 31 or high - low == 15):
			raise parser.error("floats should be 16 or 32 bit fields")
		elif not self.type in builtin_types and not self.type in parser.enums:
			raise parser.error("unknown type '%s'" % self.type);

	def ctype(self, var_name):
		if self.type == None:
			type = "uint32_t"
			val = var_name
		elif self.type == "boolean":
			type = "bool"
			val = var_name
		elif self.type == "uint" or self.type == "hex" or self.type == "a3xx_regid":
			type = "uint32_t"
			val = var_name
		elif self.type == "int":
			type = "int32_t"
			val = var_name
		elif self.type == "fixed":
			type = "float"
			val = "((int32_t)(%s * %d.0))" % (var_name, 1 << self.radix)
		elif self.type == "ufixed":
			type = "float"
			val = "((uint32_t)(%s * %d.0))" % (var_name, 1 << self.radix)
		elif self.type == "float" and self.high - self.low == 31:
			type = "float"
			val = "fui(%s)" % var_name
		elif self.type == "float" and self.high - self.low == 15:
			type = "float"
			val = "_mesa_float_to_half(%s)" % var_name
		elif self.type in [ "address", "waddress" ]:
			type = "uint64_t"
			val = var_name
		else:
			type = "enum %s" % self.type
			val = var_name

		if self.shr > 0:
			val = "(%s >> %d)" % (val, self.shr)

		return (type, val)

def tab_to(name, value):
	tab_count = (68 - (len(name) & ~7)) // 8
	if tab_count <= 0:
		tab_count = 1
	print(name + ('\t' * tab_count) + value)

def mask(low, high):
	return ((0xffffffff >> (32 - (high + 1 - low))) << low)

class Bitset(object):
	def __init__(self, name, template):
		self.name = name
		self.inline = False
		if template:
			self.fields = template.fields[:]
		else:
			self.fields = []

	def dump_pack_struct(self, prefix=None, array=None, bit_size=32):
		def field_name(prefix, name):
			if f.name:
				name = f.name.lower()
			else:
				name = prefix.lower()

			if (name in [ "double", "float", "int" ]) or not (name[0].isalpha()):
					name = "_" + name

			return name

		if not prefix:
			return
		if prefix == None:
			prefix = self.name

		value_name = "dword"
		print("struct %s {" % prefix)
		for f in self.fields:
			if f.type in [ "address", "waddress" ]:
				tab_to("    __bo_type", "bo;")
				tab_to("    uint32_t", "bo_offset;")
				if bit_size == 64:
					value_name = "qword"
				continue
			name = field_name(prefix, f.name)

			type, val = f.ctype("var")

			tab_to("    %s" % type, "%s;" % name)
		if value_name == "qword":
			tab_to("    uint64_t", "unknown;")
			tab_to("    uint64_t", "qword;")
		else:
			tab_to("    uint32_t", "unknown;")
			tab_to("    uint32_t", "dword;")
		print("};\n")

		address = None;
		for f in self.fields:
			if f.type in [ "address", "waddress" ]:
				address = f
		if array:
			print("static inline struct fd_reg_pair\npack_%s(uint32_t i, struct %s fields)\n{" %
				  (prefix, prefix));
		else:
			print("static inline struct fd_reg_pair\npack_%s(struct %s fields)\n{" %
				  (prefix, prefix));

		print("#ifndef NDEBUG")
		known_mask = 0
		for f in self.fields:
			known_mask |= mask(f.low, f.high)
			if f.type in [ "boolean", "address", "waddress" ]:
				continue
			type, val = f.ctype("fields.%s" % field_name(prefix, f.name))
			print("    assert((%-40s & 0x%08x) == 0);" % (val, 0xffffffff ^ mask(0 , f.high - f.low)))
		print("    assert((%-40s & 0x%08x) == 0);" % ("fields.unknown", known_mask))
		print("#endif\n")

		print("    return (struct fd_reg_pair) {")
		if array:
			print("        .reg = REG_%s(i)," % prefix)
		else:
			print("        .reg = REG_%s," % prefix)

		print("        .value =")
		for f in self.fields:
			if f.type in [ "address", "waddress" ]:
				continue
			else:
				type, val = f.ctype("fields.%s" % field_name(prefix, f.name))
				print("            (%-40s << %2d) |" % (val, f.low))
		print("            fields.unknown | fields.%s," % (value_name,))

		if address:
			print("        .is_address = true,")
			print("        .bo = fields.bo,")
			if f.type == "waddress":
				print("        .bo_write = true,")
			print("        .bo_offset = fields.bo_offset,")
			print("        .bo_shift = %d" % address.shr)

		print("    };\n}\n")

		if address:
			skip = ", { .reg = 0 }"
		else:
			skip = ""

		if array:
			print("#define %s(i, ...) pack_%s(i, (struct %s) { __VA_ARGS__ })%s\n" %
				  (prefix, prefix, prefix, skip))
		else:
			print("#define %s(...) pack_%s((struct %s) { __VA_ARGS__ })%s\n" %
				  (prefix, prefix, prefix, skip))


	def dump(self, prefix=None):
		if prefix == None:
			prefix = self.name
		for f in self.fields:
			if f.name:
				name = prefix + "_" + f.name
			else:
				name = prefix

			if not f.name and f.low == 0 and f.shr == 0 and not f.type in ["float", "fixed", "ufixed"]:
				pass
			elif f.type == "boolean" or (f.type == None and f.low == f.high):
				tab_to("#define %s" % name, "0x%08x" % (1 << f.low))
			else:
				tab_to("#define %s__MASK" % name, "0x%08x" % mask(f.low, f.high))
				tab_to("#define %s__SHIFT" % name, "%d" % f.low)
				type, val = f.ctype("val")

				print("static inline uint32_t %s(%s val)\n{" % (name, type))
				if f.shr > 0:
					print("\tassert(!(val & 0x%x));" % mask(0, f.shr - 1))
				print("\treturn ((%s) << %s__SHIFT) & %s__MASK;\n}" % (val, name, name))
		print()

class Array(object):
	def __init__(self, attrs, domain):
		if "name" in attrs:
			self.name = attrs["name"]
		else:
			self.name = ""
		self.domain = domain
		self.offset = int(attrs["offset"], 0)
		self.stride = int(attrs["stride"], 0)
		self.length = int(attrs["length"], 0)

	def dump(self):
		print("#define REG_%s_%s(i0) (0x%08x + 0x%x*(i0))\n" % (self.domain, self.name, self.offset, self.stride))

	def dump_pack_struct(self):
		pass

class Reg(object):
	def __init__(self, attrs, domain, array, bit_size):
		self.name = attrs["name"]
		self.domain = domain
		self.array = array
		self.offset = int(attrs["offset"], 0)
		self.type = None
		self.bit_size = bit_size

		if self.array:
			self.full_name = self.domain + "_" + self.array.name + "_" + self.name
		else:
			self.full_name = self.domain + "_" + self.name

	def dump(self):
		if self.array:
			offset = self.array.offset + self.offset
			print("static inline uint32_t REG_%s(uint32_t i0) { return 0x%08x + 0x%x*i0; }" % (self.full_name, offset, self.array.stride))
		else:
			tab_to("#define REG_%s" % self.full_name, "0x%08x" % self.offset)

		if self.bitset.inline:
			self.bitset.dump(self.full_name)
		print("")

	def dump_pack_struct(self):
		if self.bitset.inline:
			self.bitset.dump_pack_struct(self.full_name, not self.array == None, self.bit_size)


def parse_variants(attrs):
		if not "variants" in attrs:
				return None
		variant = attrs["variants"].split(",")[0]
		if "-" in variant:
			variant = variant[:variant.index("-")]

		return variant

class Parser(object):
	def __init__(self):
		self.current_array = None
		self.current_domain = None
		self.current_prefix = None
		self.current_stripe = None
		self.current_bitset = None
		self.bitsets = {}
		self.enums = {}
		self.file = []

	def error(self, message):
		parser, filename = self.stack[-1]
		return Error("%s:%d:%d: %s" % (filename, parser.CurrentLineNumber, parser.CurrentColumnNumber, message))

	def prefix(self):
		if self.current_stripe:
			return self.current_stripe + "_" + self.current_domain
		elif self.current_prefix:
			return self.current_prefix + "_" + self.current_domain
		else:
			return self.current_domain

	def parse_field(self, name, attrs):
		try:
			if "pos" in attrs:
				high = low = int(attrs["pos"], 0)
			elif "high" in attrs and "low" in attrs:
				high = int(attrs["high"], 0)
				low = int(attrs["low"], 0)
			else:
				low = 0
				high = 31

			if "type" in attrs:
				type = attrs["type"]
			else:
				type = None
	
			if "shr" in attrs:
				shr = int(attrs["shr"], 0)
			else:
				shr = 0

			b = Field(name, low, high, shr, type, self)

			if type == "fixed" or type == "ufixed":
				b.radix = int(attrs["radix"], 0)

			self.current_bitset.fields.append(b)
		except ValueError as e:
			raise self.error(e);

	def do_parse(self, filename):
		file = open(filename, "rb")
		parser = xml.parsers.expat.ParserCreate()
		self.stack.append((parser, filename))
		parser.StartElementHandler = self.start_element
		parser.EndElementHandler = self.end_element
		parser.ParseFile(file)
		self.stack.pop()
		file.close()

	def parse(self, rnn_path, filename):
		self.path = rnn_path
		self.stack = []
		self.do_parse(filename)

	def parse_reg(self, attrs, bit_size):
		if "type" in attrs and attrs["type"] in self.bitsets:
			bitset = self.bitsets[attrs["type"]]
			if bitset.inline:
				self.current_bitset = Bitset(attrs["name"], bitset)
				self.current_bitset.inline = True
			else:
				self.current_bitset = bitset
		else:
			self.current_bitset = Bitset(attrs["name"], None)
			self.current_bitset.inline = True
			if "type" in attrs:
				self.parse_field(None, attrs)

		self.current_reg = Reg(attrs, self.prefix(), self.current_array, bit_size)
		self.current_reg.bitset = self.current_bitset

		if len(self.stack) == 1:
			self.file.append(self.current_reg)

	def start_element(self, name, attrs):
		if name == "import":
			filename = attrs["file"]
			self.do_parse(os.path.join(self.path, filename))
		elif name == "domain":
			self.current_domain = attrs["name"]
			if "prefix" in attrs and attrs["prefix"] == "chip":
				self.current_prefix = parse_variants(attrs)
		elif name == "stripe":
			self.current_stripe = parse_variants(attrs)
		elif name == "enum":
			self.current_enum_value = 0
			self.current_enum = Enum(attrs["name"])
			self.enums[attrs["name"]] = self.current_enum
			if len(self.stack) == 1:
				self.file.append(self.current_enum)
		elif name == "value":
			if "value" in attrs:
				value = int(attrs["value"], 0)
			else:
				value = self.current_enum_value
			self.current_enum.values.append((attrs["name"], value))
			# self.current_enum_value = value + 1
		elif name == "reg32":
			self.parse_reg(attrs, 32)
		elif name == "reg64":
			self.parse_reg(attrs, 64)
		elif name == "array":
			self.current_array = Array(attrs, self.prefix())
			if len(self.stack) == 1:
				self.file.append(self.current_array)
		elif name == "bitset":
			self.current_bitset = Bitset(attrs["name"], None)
			if "inline" in attrs and attrs["inline"] == "yes":
				self.current_bitset.inline = True
			self.bitsets[self.current_bitset.name] = self.current_bitset
			if len(self.stack) == 1 and not self.current_bitset.inline:
				self.file.append(self.current_bitset)
		elif name == "bitfield" and self.current_bitset:
			self.parse_field(attrs["name"], attrs)

	def end_element(self, name):
		if name == "domain":
			self.current_domain = None
			self.current_prefix = None
		elif name == "stripe":
			self.current_stripe = None
		elif name == "bitset":
			self.current_bitset = None
		elif name == "reg32":
			self.current_reg = None
		elif name == "array":
			self.current_array = None;
		elif name == "enum":
			self.current_enum = None

	def dump(self):
		enums = []
		bitsets = []
		regs = []
		for e in self.file:
			if isinstance(e, Enum):
				enums.append(e)
			elif isinstance(e, Bitset):
				bitsets.append(e)
			else:
				regs.append(e)

		for e in enums + bitsets + regs:
			e.dump()

	def dump_structs(self):
		for e in self.file:
			e.dump_pack_struct()


def main():
	p = Parser()
	rnn_path = sys.argv[1]
	xml_file = sys.argv[2]
	if len(sys.argv) > 3 and sys.argv[3] == '--pack-structs':
		do_structs = True
		guard = str.replace(os.path.basename(xml_file), '.', '_').upper() + '_STRUCTS'
	else:
		do_structs = False
		guard = str.replace(os.path.basename(xml_file), '.', '_').upper()

	print("#ifndef %s\n#define %s\n" % (guard, guard))

	try:
		p.parse(rnn_path, xml_file)
	except Error as e:
		print(e)
		exit(1)

	if do_structs:
		p.dump_structs()
	else:
		p.dump()

	print("\n#endif /* %s */" % guard)

if __name__ == '__main__':
	main()



COPYRIGHT = """\
/*
 * Copyright 2017 Google
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sub license, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * IN NO EVENT SHALL VMWARE AND/OR ITS SUPPLIERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
"""

# stdlib
import argparse
from sys import stdout
from mako.template import Template

# local
import format_parser

def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument("csv")
    p.add_argument("out")
    return p.parse_args()

def get_unorm_to_srgb_map(formats):
    names = set(fmt.name for fmt in formats)

    for fmt in formats:
        if fmt.colorspace != 'srgb':
            continue;

        replacements = [
            ('SRGB', 'RGB'),
            ('SRGB', 'UNORM'),
            ('SRGB8_ALPHA8', 'RGBA'),
            ('SRGB8_ALPHA8', 'RGBA8'),
            ('SRGB_ALPHA_UNORM', 'RGBA_UNORM'),
        ]
        found_unorm_name = False
        for rep in replacements:
            if fmt.name.find(rep[0]) == -1:
                continue

            unorm_name = fmt.name.replace(rep[0], rep[1])
            if unorm_name in names:
                yield unorm_name, fmt.name
                found_unorm_name = True
                break

        # Every sRGB format MUST have a UNORM equivalent
        assert found_unorm_name

def get_intensity_to_red_map(formats):
    names = set(fmt.name for fmt in formats)

    for fmt in formats:
        if str(fmt.swizzle) != 'xxxx':
            continue

        i_name = fmt.name
        r_name = i_name.replace("_I_", "_R_")

        assert r_name in names

        yield i_name, r_name

TEMPLATE = Template(COPYRIGHT + """
#include "formats.h"
#include "util/macros.h"

/**
 * For an sRGB format, return the corresponding linear color space format.
 * For non-sRGB formats, return the format as-is.
 */
mesa_format
_mesa_get_srgb_format_linear(mesa_format format)
{
   switch (format) {
%for unorm, srgb in unorm_to_srgb_map:
   case ${srgb}:
      return ${unorm};
%endfor
   default:
      return format;
   }
}

/**
 * For an intensity format, return the corresponding red format.  For other
 * formats, return the format as-is.
 */
mesa_format
_mesa_get_intensity_format_red(mesa_format format)
{
   switch (format) {
%for i, r in intensity_to_red_map:
   case ${i}:
      return ${r};
%endfor
   default:
      return format;
   }
}
""");

def main():
    pargs = parse_args()

    formats = list(format_parser.parse(pargs.csv))

    template_env = {
        'unorm_to_srgb_map': list(get_unorm_to_srgb_map(formats)),
        'intensity_to_red_map': list(get_intensity_to_red_map(formats)),
    }

    with open(pargs.out, 'w') as f:
        f.write(TEMPLATE.render(**template_env))

if __name__ == "__main__":
    main()
